"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var object_helper_1 = require("./util/object-helper");
var types_1 = require("./util/types");
/**
 * Base Implementation of a name mangler
 */
var BaseNameMangler = (function () {
    function BaseNameMangler(compilationContext) {
        this.compilationContext = compilationContext;
    }
    Object.defineProperty(BaseNameMangler.prototype, "typeChecker", {
        get: function () {
            return this.compilationContext.typeChecker;
        },
        enumerable: true,
        configurable: true
    });
    BaseNameMangler.prototype.mangleFunctionName = function (name, argumentTypes, sourceFile) {
        var parts = [
            this.getModulePrefix(sourceFile),
            this.getFunctionName(name, argumentTypes)
        ];
        return parts.filter(function (part) { return !!part; }).join(this.separator);
    };
    BaseNameMangler.prototype.mangleMethodName = function (clazz, methodName, argumentTypes, sourceFile) {
        var parts = [
            this.getModulePrefix(sourceFile),
            this.getObjectName(clazz),
            this.getFunctionName(methodName, argumentTypes)
        ];
        return parts.filter(function (part) { return !!part; }).join(this.separator);
    };
    BaseNameMangler.prototype.mangleProperty = function (property, setter) {
        return this.mangleAccessor(this.encodeName(property.name.text), property, setter);
    };
    BaseNameMangler.prototype.mangleIndexer = function (element, setter) {
        return this.mangleAccessor(setter ? "set" : "get", element, setter);
    };
    /**
     * Returns the type code for the given parameter
     * @param parameterType the parameter to encode
     * @return {string} by default, the type code of the parameter type
     */
    BaseNameMangler.prototype.getParameterTypeCode = function (parameterType) {
        return this.typeToCode(parameterType);
    };
    BaseNameMangler.prototype.mangleAccessor = function (name, node, setter) {
        var argumentTypes = [];
        if (node.kind === ts.SyntaxKind.ElementAccessExpression) {
            argumentTypes.push(this.typeChecker.getTypeAtLocation(node.argumentExpression));
        }
        if (setter) {
            argumentTypes.push(this.typeChecker.getTypeAtLocation(node));
        }
        var classType = object_helper_1.getTypeOfParentObject(node, this.typeChecker);
        return this.mangleMethodName(classType, name, argumentTypes);
    };
    BaseNameMangler.prototype.getFunctionName = function (name, argumentTypes) {
        var _this = this;
        var parameterPostfix = argumentTypes.map(function (type) { return _this.getParameterTypeCode(type); }).join("");
        return this.encodeName(name) + parameterPostfix;
    };
    BaseNameMangler.prototype.getObjectName = function (objectType) {
        var _this = this;
        var symbol = objectType.getSymbol();
        var name = symbol.getName();
        if (objectType.objectFlags & ts.ObjectFlags.Reference) {
            var referenceType = objectType;
            if (referenceType.typeArguments && referenceType.typeArguments.length > 0) {
                name += "I";
                name += referenceType.typeArguments.map(function (typeArgument) { return _this.typeToCode(typeArgument); }).join("");
            }
        }
        return name;
    };
    BaseNameMangler.prototype.typeToCode = function (type) {
        var _this = this;
        if (type.flags & ts.TypeFlags.BooleanLike) {
            return "b";
        }
        if (type.flags & ts.TypeFlags.IntLike) {
            return "i";
        }
        if (type.flags & ts.TypeFlags.NumberLike) {
            return "d";
        }
        if (type.flags & ts.TypeFlags.Void) {
            return "v";
        }
        if (types_1.isMaybeObjectType(type)) {
            return "?" + this.typeToCode(type.getNonNullableType());
        }
        if (type.flags & ts.TypeFlags.Object) {
            var objectType = type;
            var objectName = this.encodeName(objectType.getSymbol().getName());
            var typeArgumentsPostfix = "";
            if (objectType.objectFlags & ts.ObjectFlags.Reference) {
                var typeArguments = type.typeArguments;
                typeArgumentsPostfix = "I" + (typeArguments || []).map(function (arg) { return _this.typeToCode(arg); });
            }
            return "" + objectName + typeArgumentsPostfix;
        }
        throw new Error("Unsupported runtime type " + this.typeChecker.typeToString(type));
    };
    return BaseNameMangler;
}());
exports.BaseNameMangler = BaseNameMangler;
//# sourceMappingURL=base-name-mangler.js.map