"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var fs = require("fs");
var path = require("path");
var tmp = require("tmp");
/**
 * Stores the result of a build and allows to delete the directory together with all created files
 */
var BuildDirectory = (function () {
    /**
     * Creates a new build directory that safes the intermediate files at the given path
     * @param path a path or a temp object that points to a directory
     */
    function BuildDirectory(path) {
        this.tempObjects = [];
        this.directoriesToClean = [];
        if (typeof (path) === "string") {
            this.path = path;
        }
        else {
            this.path = path.name;
            this.tempObjects.push(path);
        }
    }
    /**
     * Creates a new temporary build directory
     * @return {BuildDirectory} the build directory
     */
    BuildDirectory.createTempBuildDirectory = function () {
        return new BuildDirectory(tmp.dirSync());
    };
    /**
     * Returns a new temporary file in the build directory with the given postfix.
     * The file is deleted when the file directory is removed
     * @param postfix the postfix of the file
     * @return {string} the path to the temporary file
     */
    BuildDirectory.prototype.getTempFileName = function (postfix) {
        var file = tmp.fileSync({ dir: this.path, postfix: postfix });
        this.tempObjects.push(file);
        return file.name;
    };
    /**
     * Creates a temporary sub directory in the build directory
     * @param postfix a postfix for the name of the directory
     * @return {string} the absolute path to the sub directory
     */
    BuildDirectory.prototype.getTempSubdirectory = function (postfix) {
        var dir = tmp.dirSync({ dir: this.path, postfix: postfix });
        this.tempObjects.push(dir);
        this.directoriesToClean.push(dir.name);
        return dir.name;
    };
    /**
     * Deletes all files and sub directories in the build directory and finally deletes the build directory itself.
     * Methods called on the build directory object will fail thereafter.
     */
    BuildDirectory.prototype.remove = function () {
        for (var _i = 0, _a = this.directoriesToClean; _i < _a.length; _i++) {
            var dir = _a[_i];
            this.deleteDirectory(dir);
        }
        for (var i = this.tempObjects.length - 1; i >= 0; --i) {
            this.tempObjects[i].removeCallback();
        }
    };
    BuildDirectory.prototype.deleteDirectory = function (dir) {
        for (var _i = 0, _a = fs.readdirSync(dir); _i < _a.length; _i++) {
            var file = _a[_i];
            var filePath = path.join(dir, file);
            if (fs.lstatSync(filePath).isDirectory()) {
                this.deleteDirectory(filePath);
            }
            else {
                fs.unlinkSync(filePath);
            }
        }
    };
    return BuildDirectory;
}());
exports.BuildDirectory = BuildDirectory;
//# sourceMappingURL=build-directory.js.map