"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var ts = require("typescript");
var types_1 = require("./util/types");
var address_lvalue_1 = require("./value/address-lvalue");
var primitive_1 = require("./value/primitive");
var resolved_function_reference_1 = require("./value/resolved-function-reference");
var speedy_js_class_reference_1 = require("./value/speedy-js-class-reference");
/**
 * Defines the extension methods / default implementations that do not depend on a particular code generation context implementation
 */
var CodeGenerationContextMixin = (function () {
    function CodeGenerationContextMixin() {
    }
    CodeGenerationContextMixin.prototype.generateValue = function (node) {
        var result = this.generate(node);
        assert(result, "Generator for node of kind " + ts.SyntaxKind[node.kind] + " returned no value but caller expected value");
        return result;
    };
    CodeGenerationContextMixin.prototype.generateChildren = function (node) {
        var _this = this;
        ts.forEachChild(node, function (child) {
            _this.generate(child);
        });
    };
    CodeGenerationContextMixin.prototype.assignValue = function (target, value) {
        if (target.isAssignable()) {
            target.generateAssignmentIR(value, this);
        }
        else {
            throw new Error("Assignment to readonly value " + target);
        }
    };
    CodeGenerationContextMixin.prototype.value = function (value, type) {
        var symbol = type.getSymbol();
        if (type.flags & (ts.TypeFlags.BooleanLike | ts.TypeFlags.NumberLike | ts.TypeFlags.IntLike)) {
            return new primitive_1.Primitive(value, type);
        }
        if (symbol) {
            if (symbol.flags & ts.SymbolFlags.Function) {
                var signatures = this.typeChecker.getSignaturesOfType(type, ts.SignatureKind.Call);
                assert(signatures.length === 1, "No function type found or function is overloaded und should therefore not be dereferenced");
                return resolved_function_reference_1.ResolvedFunctionReference.createForSignature(value, signatures[0], this);
            }
            if (symbol.flags & ts.SymbolFlags.Method) {
                // TODO Objekt erstellen und dann methode
                // Requires special return object that contains the method function pointer and as
                // well the object reference ptr
                throw new Error("Returning methods is not yet supported");
            }
        }
        if (types_1.isMaybeObjectType(type)) {
            type = type.getNonNullableType();
        }
        if (type.flags & ts.TypeFlags.Object) {
            var classReference = this.resolveClass(type);
            if (classReference) {
                return classReference.objectFor(new address_lvalue_1.AddressLValue(value, type), type, this);
            }
        }
        throw Error("Unable to convert llvm value of type " + this.typeChecker.typeToString(type) + " to Value object.");
    };
    CodeGenerationContextMixin.prototype.resolveClass = function (type, symbol) {
        if (symbol === void 0) { symbol = type.getSymbol(); }
        if (this.scope.hasClass(symbol)) {
            return this.scope.getClass(symbol);
        }
        if (symbol.flags & ts.SymbolFlags.Class && isClassDefined(type)) {
            var reference = speedy_js_class_reference_1.SpeedyJSClassReference.create(type, this);
            this.scope.addClass(symbol, reference);
            return reference;
        }
        return undefined;
    };
    return CodeGenerationContextMixin;
}());
exports.CodeGenerationContextMixin = CodeGenerationContextMixin;
function isClassDefined(type) {
    return type.getProperties().every(function (property) {
        if (property.flags & ts.SymbolFlags.Property) {
            // simple property / field
            return true;
        }
        var declarations = property.getDeclarations() || [];
        if (declarations.length === 0) {
            return false;
        }
        var declaration = declarations[0];
        return declaration.kind === ts.SyntaxKind.MethodDeclaration ||
            (declaration.kind === ts.SyntaxKind.Constructor && !!declaration.body);
    });
}
/**
 * Applies the {@link CodeGenerationContextMixin} to the given type
 * @param derivedCtor the type for which the
 */
function applyCodeGenerationContextMixin(derivedCtor) {
    Object.getOwnPropertyNames(CodeGenerationContextMixin.prototype).forEach(function (name) {
        derivedCtor.prototype[name] = CodeGenerationContextMixin.prototype[name];
    });
}
exports.applyCodeGenerationContextMixin = applyCodeGenerationContextMixin;
//# sourceMappingURL=code-generation-context-mixin.js.map