"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var llvm = require("llvm-node");
var ts = require("typescript");
var code_generation_diagnostic_1 = require("../../code-generation-diagnostic");
var types_1 = require("../util/types");
var math_object_reference_1 = require("../value/math-object-reference");
var primitive_1 = require("../value/primitive");
function isAssignment(operatorToken) {
    return operatorToken.kind === ts.SyntaxKind.EqualsToken ||
        operatorToken.kind === ts.SyntaxKind.PlusEqualsToken ||
        operatorToken.kind === ts.SyntaxKind.MinusEqualsToken ||
        operatorToken.kind === ts.SyntaxKind.AsteriskAsteriskEqualsToken ||
        operatorToken.kind === ts.SyntaxKind.AsteriskEqualsToken ||
        operatorToken.kind === ts.SyntaxKind.SlashEqualsToken ||
        operatorToken.kind === ts.SyntaxKind.PercentEqualsToken ||
        operatorToken.kind === ts.SyntaxKind.AmpersandEqualsToken ||
        operatorToken.kind === ts.SyntaxKind.BarEqualsToken ||
        operatorToken.kind === ts.SyntaxKind.CaretEqualsToken ||
        operatorToken.kind === ts.SyntaxKind.LessThanLessThanEqualsToken ||
        operatorToken.kind === ts.SyntaxKind.GreaterThanGreaterThanGreaterThanEqualsToken ||
        operatorToken.kind === ts.SyntaxKind.GreaterThanGreaterThanEqualsToken;
}
/**
 * Code Generator for binary expressions, e.g. 5+3 but also x = 3, or x += 3
 */
var BinaryExpressionCodeGenerator = (function () {
    function BinaryExpressionCodeGenerator() {
        this.syntaxKind = ts.SyntaxKind.BinaryExpression;
    }
    BinaryExpressionCodeGenerator.prototype.generate = function (binaryExpression, context) {
        var leftType = context.typeChecker.getTypeAtLocation(binaryExpression.left);
        var rightType = context.typeChecker.getTypeAtLocation(binaryExpression.right);
        var resultType = context.typeChecker.getTypeAtLocation(binaryExpression);
        var result;
        var resultValue;
        switch (binaryExpression.operatorToken.kind) {
            // 10.12 & 0
            case ts.SyntaxKind.AmpersandToken:
            case ts.SyntaxKind.AmpersandEqualsToken: {
                var leftInt = primitive_1.Primitive.toInt32(context.generateValue(binaryExpression.left), leftType, resultType, context).generateIR();
                var rightInt = primitive_1.Primitive.toInt32(context.generateValue(binaryExpression.right), rightType, resultType, context).generateIR();
                result = context.builder.createAnd(leftInt, rightInt, "and");
                break;
            }
            // a || b
            case ts.SyntaxKind.AmpersandAmpersandToken: {
                var lhs = context.generateValue(binaryExpression.left).generateIR(context);
                var lhsAsBool = primitive_1.Primitive.toBoolean(lhs, leftType, context);
                var lhsBlock = context.builder.getInsertBlock();
                var rhsBlock = llvm.BasicBlock.create(context.llvmContext, "land.lhs.true");
                var end = llvm.BasicBlock.create(context.llvmContext, "land.end");
                context.builder.createCondBr(lhsAsBool, rhsBlock, end);
                context.scope.enclosingFunction.addBasicBlock(rhsBlock);
                context.builder.setInsertionPoint(rhsBlock);
                var right = context.generateValue(binaryExpression.right).generateIR(context);
                context.builder.createBr(end);
                rhsBlock = context.builder.getInsertBlock();
                context.scope.enclosingFunction.addBasicBlock(end);
                context.builder.setInsertionPoint(end);
                var phi = context.builder.createPhi(types_1.toLLVMType(resultType, context), 2, "land");
                phi.addIncoming(lhs, lhsBlock);
                phi.addIncoming(right, rhsBlock);
                result = phi;
                break;
            }
            // a * b
            case ts.SyntaxKind.AsteriskToken:
            case ts.SyntaxKind.AsteriskEqualsToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var leftIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var rightIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                if (unifiedType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createMul(leftIr, rightIr, "mul");
                }
                else if (unifiedType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFMul(leftIr, rightIr, "mul");
                }
                break;
            }
            // a ** b
            case ts.SyntaxKind.AsteriskAsteriskToken:
            case ts.SyntaxKind.AsteriskAsteriskEqualsToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var left = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context);
                var right = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context);
                if (unifiedType.flags & (ts.TypeFlags.IntLike | ts.TypeFlags.NumberLike)) {
                    result = math_object_reference_1.MathObjectReference.pow(left, leftType, right, rightType, resultType, context).generateIR(context);
                }
                break;
            }
            // a | b
            case ts.SyntaxKind.BarToken:
            case ts.SyntaxKind.BarEqualsToken: {
                var intType = resultType;
                var left = context.generateValue(binaryExpression.left);
                var lhsIntValue = primitive_1.Primitive.toInt32(left, leftType, intType, context).generateIR();
                var right = context.generateValue(binaryExpression.right);
                var rhsIntValue = primitive_1.Primitive.toInt32(right, rightType, intType, context).generateIR();
                result = context.builder.createOr(lhsIntValue, rhsIntValue, "or");
                break;
            }
            // a || b
            case ts.SyntaxKind.BarBarToken: {
                var lhs = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var lhsBlock = context.builder.getInsertBlock();
                var lhsAsBool = primitive_1.Primitive.toBoolean(lhs, leftType, context);
                var rhsBlock = llvm.BasicBlock.create(context.llvmContext, "lor.lhs.false");
                var lorEnd = llvm.BasicBlock.create(context.llvmContext, "lor.end");
                context.builder.createCondBr(lhsAsBool, lorEnd, rhsBlock);
                context.scope.enclosingFunction.addBasicBlock(rhsBlock);
                context.builder.setInsertionPoint(rhsBlock);
                var rhs = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                context.builder.createBr(lorEnd);
                rhsBlock = context.builder.getInsertBlock();
                context.scope.enclosingFunction.addBasicBlock(lorEnd);
                context.builder.setInsertionPoint(lorEnd);
                var phi = context.builder.createPhi(types_1.toLLVMType(resultType, context), 2, "lor");
                phi.addIncoming(lhs, lhsBlock);
                phi.addIncoming(rhs, rhsBlock);
                result = phi;
                break;
            }
            // a ^ b
            case ts.SyntaxKind.CaretEqualsToken:
            case ts.SyntaxKind.CaretToken: {
                var leftInt = primitive_1.Primitive.toInt32(context.generateValue(binaryExpression.left), leftType, resultType, context).generateIR();
                var rightInt = primitive_1.Primitive.toInt32(context.generateValue(binaryExpression.right), rightType, resultType, context).generateIR();
                result = context.builder.createXor(leftInt, rightInt, "xor");
                break;
            }
            // a, b, c
            case ts.SyntaxKind.CommaToken: {
                context.generateValue(binaryExpression.left).generateIR(context);
                result = context.generateValue(binaryExpression.right).generateIR(context);
                break;
            }
            // a === b
            case ts.SyntaxKind.EqualsEqualsEqualsToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var leftIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var rightIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                // tslint:disable-next-line: max-line-length
                if (unifiedType.flags & (ts.TypeFlags.IntLike | ts.TypeFlags.BooleanLike | ts.TypeFlags.Object | ts.TypeFlags.Undefined) || types_1.isMaybeObjectType(unifiedType)) {
                    result = context.builder.createICmpEQ(leftIr, rightIr, "cmpEQ");
                }
                else if (unifiedType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFCmpOEQ(leftIr, rightIr, "cmpEQ");
                }
                break;
            }
            // !==
            case ts.SyntaxKind.ExclamationEqualsEqualsToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var leftIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var rightIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                // tslint:disable-next-line: max-line-length
                if (unifiedType.flags & (ts.TypeFlags.IntLike | ts.TypeFlags.BooleanLike | ts.TypeFlags.Object | ts.TypeFlags.Undefined) || types_1.isMaybeObjectType(unifiedType)) {
                    result = context.builder.createICmpNE(leftIr, rightIr, "cmpNE");
                }
                else if (leftType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFCmpONE(leftIr, rightIr, "cmpNE");
                }
                break;
            }
            // a > b
            case ts.SyntaxKind.GreaterThanToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var leftIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var rightIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                if (unifiedType.flags & ts.TypeFlags.BooleanLike) {
                    var leftInt = context.builder.createZExt(leftIr, llvm.Type.getInt32Ty(context.llvmContext));
                    var rightInt = context.builder.createZExt(rightIr, llvm.Type.getInt32Ty(context.llvmContext));
                    result = context.builder.createICmpSGT(leftInt, rightInt, "cmpGT");
                }
                else if (unifiedType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createICmpSGT(leftIr, rightIr, "cmpGT");
                }
                else if (unifiedType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFCmpOGT(leftIr, rightIr, "cmpGT");
                }
                break;
            }
            // a >= b
            case ts.SyntaxKind.GreaterThanEqualsToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var leftIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var rightIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                if (unifiedType.flags & ts.TypeFlags.BooleanLike) {
                    var leftInt = context.builder.createZExt(leftIr, llvm.Type.getInt32Ty(context.llvmContext));
                    var rightInt = context.builder.createZExt(rightIr, llvm.Type.getInt32Ty(context.llvmContext));
                    result = context.builder.createICmpSGE(leftInt, rightInt, "cmpGE");
                }
                else if (unifiedType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createICmpSGE(leftIr, rightIr, "cmpGE");
                }
                else if (unifiedType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFCmpOGE(leftIr, rightIr, "cmpGE");
                }
                break;
            }
            // a >> b
            case ts.SyntaxKind.GreaterThanGreaterThanToken:
            case ts.SyntaxKind.GreaterThanGreaterThanEqualsToken: {
                var leftInt = primitive_1.Primitive.toInt32(context.generateValue(binaryExpression.left), leftType, resultType, context).generateIR();
                var rightInt = primitive_1.Primitive.toInt32(context.generateValue(binaryExpression.right), rightType, resultType, context).generateIR();
                // mask not needed, less than 32 guaranteed by wasm
                result = context.builder.createAShr(leftInt, rightInt, "ashr");
                break;
            }
            // a >>> b
            case ts.SyntaxKind.GreaterThanGreaterThanGreaterThanEqualsToken:
            case ts.SyntaxKind.GreaterThanGreaterThanGreaterThanToken: {
                var leftInt = primitive_1.Primitive.toInt32(context.generateValue(binaryExpression.left), leftType, resultType, context).generateIR();
                var rightInt = primitive_1.Primitive.toInt32(context.generateValue(binaryExpression.right), rightType, resultType, context).generateIR();
                // mask not needed, less than 32 guaranteed by wasm
                result = context.builder.createLShr(leftInt, rightInt, "lshr");
                break;
            }
            // a < b
            case ts.SyntaxKind.LessThanToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var leftIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var rightIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                if (unifiedType.flags & ts.TypeFlags.BooleanLike) {
                    var leftInt = context.builder.createZExt(leftIr, llvm.Type.getInt32Ty(context.llvmContext));
                    var rightInt = context.builder.createZExt(rightIr, llvm.Type.getInt32Ty(context.llvmContext));
                    result = context.builder.createICmpSLT(leftInt, rightInt, "cmpLT");
                }
                else if (unifiedType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createICmpSLT(leftIr, rightIr, "cmpLT");
                }
                else if (unifiedType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFCmpOLT(leftIr, rightIr, "cmpLT");
                }
                break;
            }
            // a <= b
            case ts.SyntaxKind.LessThanEqualsToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var leftIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var rightIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                if (unifiedType.flags & ts.TypeFlags.BooleanLike) {
                    var leftInt = context.builder.createZExt(leftIr, llvm.Type.getInt32Ty(context.llvmContext));
                    var rightInt = context.builder.createZExt(rightIr, llvm.Type.getInt32Ty(context.llvmContext));
                    result = context.builder.createICmpSLE(leftInt, rightInt, "cmpLE");
                }
                else if (unifiedType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createICmpSLE(leftIr, rightIr, "cmpLE");
                }
                else if (unifiedType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFCmpOLE(leftIr, rightIr, "cmpLE");
                }
                break;
            }
            // a << b
            case ts.SyntaxKind.LessThanLessThanToken:
            case ts.SyntaxKind.LessThanLessThanEqualsToken: {
                var leftInt = primitive_1.Primitive.toInt32(context.generateValue(binaryExpression.left), leftType, resultType, context).generateIR();
                var rightInt = primitive_1.Primitive.toInt32(context.generateValue(binaryExpression.right), rightType, resultType, context).generateIR();
                // mask not needed, less than 32 guaranteed by wasm
                result = context.builder.createShl(leftInt, rightInt, "shl");
                break;
            }
            // a - b
            case ts.SyntaxKind.MinusEqualsToken:
            case ts.SyntaxKind.MinusToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var leftIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var rightIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                if (unifiedType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createSub(leftIr, rightIr, "sub");
                }
                else if (unifiedType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFSub(leftIr, rightIr, "fsub");
                }
                break;
            }
            // a % b
            case ts.SyntaxKind.PercentToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var leftIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var rightIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                if (unifiedType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createSRem(leftIr, rightIr, "srem");
                }
                else if (unifiedType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFRem(leftIr, rightIr, "frem");
                }
                break;
            }
            // a + b
            case ts.SyntaxKind.PlusEqualsToken:
            case ts.SyntaxKind.PlusToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var leftIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var rightIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                if (unifiedType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createAdd(leftIr, rightIr, "add");
                }
                else if (unifiedType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFAdd(leftIr, rightIr, "fadd");
                }
                break;
            }
            // a / b
            case ts.SyntaxKind.SlashEqualsToken:
            case ts.SyntaxKind.SlashToken: {
                var unifiedType = BinaryExpressionCodeGenerator.unifyTypes(binaryExpression, context);
                var leftIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.left, binaryExpression, context).generateIR(context);
                var rightIr = BinaryExpressionCodeGenerator.toUnified(binaryExpression.right, binaryExpression, context).generateIR(context);
                if (unifiedType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createSDiv(leftIr, rightIr, "sdiv");
                }
                else if (unifiedType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFDiv(leftIr, rightIr, "div");
                }
                break;
            }
            // a = b
            case ts.SyntaxKind.FirstAssignment:
                resultValue = context.generateValue(binaryExpression.right).castImplicit(leftType, context);
                if (!resultValue) {
                    BinaryExpressionCodeGenerator.throwUnsupportedImplicitCast(binaryExpression, context);
                }
        }
        if (result) {
            resultValue = context.value(result, resultType);
        }
        if (!resultValue) {
            throw code_generation_diagnostic_1.CodeGenerationDiagnostic.unsupportedBinaryOperation(binaryExpression, context.typeChecker.typeToString(leftType), context.typeChecker.typeToString(leftType));
        }
        if (isAssignment(binaryExpression.operatorToken)) {
            context.assignValue(context.generateValue(binaryExpression.left), resultValue);
        }
        return resultValue;
    };
    BinaryExpressionCodeGenerator.toUnified = function (node, binaryExpression, context) {
        var unifiedType = this.unifyTypes(binaryExpression, context);
        var casted = context.generateValue(node).castImplicit(unifiedType, context);
        if (!casted) {
            return this.throwUnsupportedImplicitCast(binaryExpression, context);
        }
        return casted;
    };
    BinaryExpressionCodeGenerator.unifyTypes = function (binaryExpression, context) {
        function unify(left, right) {
            if (left === right) {
                return left;
            }
            // cast int to number
            if (left.flags & ts.TypeFlags.IntLike && right.flags & (ts.TypeFlags.NumberLike ^ ts.TypeFlags.IntLike)) {
                return right;
            }
            if ((left.flags & ts.TypeFlags.Object || types_1.isMaybeObjectType(left)) && right.flags & ts.TypeFlags.Undefined) {
                return left;
            }
            return undefined;
        }
        var leftType = context.typeChecker.getTypeAtLocation(binaryExpression.left);
        var rightType = context.typeChecker.getTypeAtLocation(binaryExpression.right);
        var unified = unify(leftType, rightType) || unify(rightType, leftType);
        if (unified) {
            return unified;
        }
        return this.throwUnsupportedImplicitCast(binaryExpression, context);
    };
    BinaryExpressionCodeGenerator.throwUnsupportedImplicitCast = function (binaryExpression, context) {
        var leftType = context.typeChecker.getTypeAtLocation(binaryExpression.left);
        var rightType = context.typeChecker.getTypeAtLocation(binaryExpression.right);
        throw code_generation_diagnostic_1.CodeGenerationDiagnostic.unsupportedImplicitCastOfBinaryExpressionOperands(binaryExpression, context.typeChecker.typeToString(leftType), context.typeChecker.typeToString(rightType));
    };
    return BinaryExpressionCodeGenerator;
}());
exports.default = BinaryExpressionCodeGenerator;
//# sourceMappingURL=binary-expression-code-generator.js.map