"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var llvm = require("llvm-node");
var ts = require("typescript");
var code_generation_diagnostic_1 = require("../../code-generation-diagnostic");
var primitive_1 = require("../value/primitive");
var PrefixUnaryExpressionCodeGenerator = (function () {
    function PrefixUnaryExpressionCodeGenerator() {
        this.syntaxKind = ts.SyntaxKind.PrefixUnaryExpression;
    }
    PrefixUnaryExpressionCodeGenerator.prototype.generate = function (node, context) {
        var left = context.generateValue(node.operand);
        var operandType = context.typeChecker.getTypeAtLocation(node.operand);
        var resultType = context.typeChecker.getTypeAtLocation(node);
        var result;
        switch (node.operator) {
            case ts.SyntaxKind.ExclamationToken:
                var booleanValue = primitive_1.Primitive.toBoolean(left, operandType, context);
                result = context.builder.createNot(booleanValue, "not");
                break;
            case ts.SyntaxKind.MinusToken:
                if (operandType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createNeg(left.generateIR(context), "neg");
                }
                else if (operandType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFNeg(left.generateIR(context), "neg");
                }
                break;
            case ts.SyntaxKind.MinusMinusToken:
                if (operandType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createSub(left.generateIR(context), llvm.ConstantInt.get(context.llvmContext, 1), "sub");
                }
                else if (operandType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFSub(left.generateIR(context), llvm.ConstantFP.get(context.llvmContext, 1.0), "fsub");
                }
                break;
            case ts.SyntaxKind.PlusToken:
                if (operandType.flags & ts.TypeFlags.IntLike || operandType.flags & ts.TypeFlags.NumberLike || operandType.flags & ts.TypeFlags.BooleanLike) {
                    var castedToResultType = left.castImplicit(resultType, context);
                    result = castedToResultType ? castedToResultType.generateIR(context) : undefined;
                }
                break;
            case ts.SyntaxKind.PlusPlusToken:
                if (operandType.flags & ts.TypeFlags.IntLike) {
                    result = context.builder.createAdd(left.generateIR(context), llvm.ConstantInt.get(context.llvmContext, 1), "add");
                }
                else if (operandType.flags & ts.TypeFlags.NumberLike) {
                    result = context.builder.createFAdd(left.generateIR(context), llvm.ConstantFP.get(context.llvmContext, 1.0), "fadd");
                }
                break;
            case ts.SyntaxKind.TildeToken:
                var intValue = primitive_1.Primitive.toInt32(left, operandType, resultType, context).generateIR();
                result = context.builder.createNot(intValue, "not");
                break;
        }
        if (!result) {
            throw code_generation_diagnostic_1.CodeGenerationDiagnostic.unsupportedUnaryOperation(node, context.typeChecker.typeToString(operandType));
        }
        var resultValue = context.value(result, resultType);
        if (node.operator === ts.SyntaxKind.PlusPlusToken || node.operator === ts.SyntaxKind.MinusMinusToken) {
            context.assignValue(left, resultValue);
        }
        return resultValue;
    };
    return PrefixUnaryExpressionCodeGenerator;
}());
exports.default = PrefixUnaryExpressionCodeGenerator;
//# sourceMappingURL=prefix-unary-expression-code-generator.js.map