"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var debug = require("debug");
var fs = require("fs");
var llvm = require("llvm-node");
var default_code_generation_context_1 = require("./default-code-generation-context");
var array_class_reference_1 = require("./value/array-class-reference");
var math_class_reference_1 = require("./value/math-class-reference");
var primitive_1 = require("./value/primitive");
var unresolved_function_reference_1 = require("./value/unresolved-function-reference");
var log = debug("DefaultLLVMEmitContextFactory");
/**
 * Default code generation factory that loads the code generators 'code generators' directory.
 */
var DefaultCodeGenerationContextFactory = (function () {
    function DefaultCodeGenerationContextFactory(fallbackCodeGenerator) {
        this.fallbackCodeGenerator = fallbackCodeGenerator;
    }
    DefaultCodeGenerationContextFactory.prototype.createContext = function (compilationContext, module) {
        var context = new default_code_generation_context_1.DefaultCodeGenerationContext(compilationContext, module);
        var codeGenerators = this.getCodeGenerators();
        for (var _i = 0, codeGenerators_1 = codeGenerators; _i < codeGenerators_1.length; _i++) {
            var codeGenerator = codeGenerators_1[_i];
            context.registerCodeGenerator(codeGenerator);
        }
        context.setFallbackCodeGenerator(this.fallbackCodeGenerator);
        this.registerGlobals(context);
        return context;
    };
    DefaultCodeGenerationContextFactory.prototype.registerGlobals = function (context) {
        var builtins = context.compilationContext.builtIns;
        var arraySymbol = builtins.get("Array");
        if (arraySymbol) {
            var arrayClassReference = array_class_reference_1.ArrayClassReference.create(arraySymbol, context);
            context.scope.addClass(arraySymbol, arrayClassReference);
            context.scope.addClass(builtins.get("ArrayConstructor"), arrayClassReference);
        }
        var mathSymbol = builtins.get("Math");
        if (mathSymbol) {
            var mathClassReference = math_class_reference_1.MathClassReference.create(mathSymbol, context);
            context.scope.addClass(mathSymbol, mathClassReference);
            context.scope.addVariable(mathSymbol, mathClassReference.createGlobalVariable(mathSymbol, context));
        }
        var nan = builtins.get("NaN");
        if (nan) {
            var nanValue = llvm.ConstantFP.getNaN(llvm.Type.getDoubleTy(context.llvmContext));
            context.scope.addVariable(nan, new primitive_1.Primitive(nanValue, context.typeChecker.getTypeAtLocation(nan.valueDeclaration)));
        }
        var isNanSymbol = builtins.get("isNaN");
        if (isNanSymbol) {
            context.scope.addFunction(isNanSymbol, unresolved_function_reference_1.UnresolvedFunctionReference.createRuntimeFunction([
                context.typeChecker.getSignatureFromDeclaration(isNanSymbol.valueDeclaration)
            ], context));
        }
    };
    DefaultCodeGenerationContextFactory.prototype.getCodeGenerators = function () {
        if (!this.codeGenerators) {
            this.codeGenerators = this.loadCodeGenerators();
        }
        return this.codeGenerators;
    };
    DefaultCodeGenerationContextFactory.prototype.loadCodeGenerators = function () {
        var fileNames = fs.readdirSync(__dirname + "/code-generators");
        var codeGeneratorFiles = fileNames.filter(function (file) { return file.endsWith("code-generator.js"); });
        return codeGeneratorFiles.map(function (codeGeneratorFile) {
            var codeGeneratorModule = require(__dirname + "/code-generators/" + codeGeneratorFile);
            assert(codeGeneratorModule && codeGeneratorModule.default, "Code generator module " + codeGeneratorFile + " lacks mandatory default export");
            var constructorFunction = codeGeneratorModule.default;
            assert(constructorFunction instanceof Function, "Default export of code generator module " + codeGeneratorFile + " is not a constructor");
            // tslint:disable-next-line: max-line-length
            assert(constructorFunction.prototype.generate, "Default exported object from module " + codeGeneratorFile + " is not a CodeGenerator as it lacks the generate method.");
            log("Dynamically loaded code generator %s", codeGeneratorFile);
            return new constructorFunction();
        });
    };
    return DefaultCodeGenerationContextFactory;
}());
exports.DefaultCodeGenerationContextFactory = DefaultCodeGenerationContextFactory;
//# sourceMappingURL=default-code-generation-context-factory.js.map