"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var debug = require("debug");
var llvm = require("llvm-node");
var ts = require("typescript");
var code_generation_context_mixin_1 = require("./code-generation-context-mixin");
var scope_1 = require("./scope");
var log = debug("code-generation/default-code-generation-context");
/**
 * Default implementation of the code generation context
 */
var DefaultCodeGenerationContext = (function () {
    function DefaultCodeGenerationContext(compilationContext, module, rootScope, codeGenerators, entryFunctions, fallbackCodeGenerator) {
        if (rootScope === void 0) { rootScope = new scope_1.Scope(); }
        if (codeGenerators === void 0) { codeGenerators = new Map(); }
        if (entryFunctions === void 0) { entryFunctions = new Set(); }
        this.compilationContext = compilationContext;
        this.module = module;
        this.rootScope = rootScope;
        this.codeGenerators = codeGenerators;
        this.entryFunctions = entryFunctions;
        this.fallbackCodeGenerator = fallbackCodeGenerator;
        this.requiresGc = false;
        this.builder = new llvm.IRBuilder(this.compilationContext.llvmContext);
        this.scope = rootScope;
    }
    Object.defineProperty(DefaultCodeGenerationContext.prototype, "llvmContext", {
        get: function () {
            return this.compilationContext.llvmContext;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DefaultCodeGenerationContext.prototype, "typeChecker", {
        get: function () {
            return this.compilationContext.typeChecker;
        },
        enumerable: true,
        configurable: true
    });
    DefaultCodeGenerationContext.prototype.createChildContext = function () {
        return new DefaultCodeGenerationContext(this.compilationContext, this.module, this.rootScope, this.codeGenerators, this.entryFunctions, this.fallbackCodeGenerator);
    };
    DefaultCodeGenerationContext.prototype.generate = function (node) {
        log("Generate node " + ts.SyntaxKind[node.kind]);
        var codeGenerator = this.getCodeGenerator(node);
        return codeGenerator.generate(node, this);
    };
    DefaultCodeGenerationContext.prototype.registerCodeGenerator = function (codeGenerator) {
        assert(codeGenerator);
        var syntaxKind = codeGenerator.syntaxKind;
        assert(syntaxKind, "Code Generator returned undefined as syntax kind");
        assert(!this.codeGenerators.has(syntaxKind), "An other Code Generator is already registered for the syntax kind " + ts.SyntaxKind[syntaxKind]);
        log("Register Code Generator for syntax kind " + ts.SyntaxKind[syntaxKind]);
        this.codeGenerators.set(syntaxKind, codeGenerator);
    };
    DefaultCodeGenerationContext.prototype.setFallbackCodeGenerator = function (fallbackCodeGenerator) {
        this.fallbackCodeGenerator = fallbackCodeGenerator;
    };
    DefaultCodeGenerationContext.prototype.addEntryFunction = function (name) {
        assert(name, "Name is required");
        this.entryFunctions.add(name);
    };
    DefaultCodeGenerationContext.prototype.getEntryFunctionNames = function () {
        return Array.from(this.entryFunctions.values());
    };
    DefaultCodeGenerationContext.prototype.enterChildScope = function (fn) {
        this.scope = this.scope.enterChild(fn);
        return this.scope;
    };
    DefaultCodeGenerationContext.prototype.leaveChildScope = function () {
        var child = this.scope;
        this.scope = this.scope.exitChild();
        return child;
    };
    DefaultCodeGenerationContext.prototype.getCodeGenerator = function (node) {
        var codeGenerator = this.codeGenerators.get(node.kind) || this.fallbackCodeGenerator;
        assert(codeGenerator, "No Code Generator registered for syntax kind " + ts.SyntaxKind[node.kind] + " nor is a fallback code generator defined");
        return codeGenerator;
    };
    return DefaultCodeGenerationContext;
}());
exports.DefaultCodeGenerationContext = DefaultCodeGenerationContext;
code_generation_context_mixin_1.applyCodeGenerationContextMixin(DefaultCodeGenerationContext);
//# sourceMappingURL=default-code-generation-context.js.map