"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var ts = require("typescript");
var types_1 = require("../util/types");
var per_file_wasm_loader_emit_helper_1 = require("./per-file-wasm-loader-emit-helper");
/**
 * Inserts a wasm module loader function, the wasm byte code and rewrites the entry functions to call the wasm module.
 */
var PerFileCodeGeneratorSourceFileRewriter = (function () {
    function PerFileCodeGeneratorSourceFileRewriter(context) {
        this.context = context;
        this.wastMetaData = {};
    }
    PerFileCodeGeneratorSourceFileRewriter.prototype.setWastMetaData = function (metadata) {
        this.wastMetaData = metadata;
    };
    PerFileCodeGeneratorSourceFileRewriter.prototype.setWasmUrl = function (wasmUrl) {
        this.wasmUrl = wasmUrl;
    };
    /**
     * Generates a new function declaration with an equal signature but replaced body. The body looks like
     * @code
     * const instance = await loadWasmModule();
     * const result = instance.exports.fn.apply(undefined, arguments); // arguments are potentially cased
     *
     * speedyJsGc(); // only if not disableHeapNukeOnExit is set
     *
     * return result; // result is potentially casted
     *
     * fn is replaced with the passed in name
     * @param name the name of the function in the compilation
     * @param functionDeclaration the function declaration to transform
     * @return the new function declaration that acts as proxy / facade for a wasm function
     */
    PerFileCodeGeneratorSourceFileRewriter.prototype.rewriteEntryFunction = function (name, functionDeclaration) {
        var _this = this;
        this.loadWasmFunctionIdentifier = this.loadWasmFunctionIdentifier || ts.createUniqueName("loadWasmModule");
        var signature = this.context.typeChecker.getSignatureFromDeclaration(functionDeclaration);
        var bodyStatements = [];
        // types = { ... }
        var argumentTypes = signature.declaration.parameters.map(function (parameter) { return _this.context.typeChecker.getTypeAtLocation(parameter); });
        var serializedTypes = this.serializeArgumentAndReturnTypes(argumentTypes, signature.getReturnType());
        var typesIdentifier = ts.createUniqueName("types");
        var typesDeclaration = ts.createVariableStatement(undefined, [ts.createVariableDeclaration(typesIdentifier, undefined, serializedTypes)]);
        bodyStatements.push(typesDeclaration);
        var argumentObjects;
        // argumentObjects = new Map();
        if (argumentTypes.findIndex(function (argumentType) { return (argumentType.flags & ts.TypeFlags.Object) !== 0 || types_1.isMaybeObjectType(argumentType); }) !== -1) {
            argumentObjects = ts.createUniqueName("argumentObjects");
            var argumentObjectsMap = ts.createNew(ts.createIdentifier("Map"), undefined, []);
            bodyStatements.push(ts.createVariableStatement(undefined, [ts.createVariableDeclaration(argumentObjects, undefined, argumentObjectsMap)]));
        }
        else {
            argumentObjects = ts.createIdentifier("undefined");
        }
        // const result = instance.exports.fn(args)
        var instanceIdentifier = ts.createUniqueName("instance");
        var wasmExports = ts.createPropertyAccess(instanceIdentifier, "exports");
        var targetFunction = ts.createPropertyAccess(wasmExports, name);
        var args = signature.declaration.parameters.map(function (parameter) { return _this.castToWasm(parameter, _this.loadWasmFunctionIdentifier, typesIdentifier, argumentObjects); });
        var functionCall = ts.createCall(targetFunction, [], args);
        var castedResult = this.castToJs(functionCall, signature.getReturnType(), this.loadWasmFunctionIdentifier, typesIdentifier);
        var resultIdentifier = ts.createUniqueName("result");
        var resultVariable = ts.createVariableDeclaration(resultIdentifier, undefined, castedResult);
        bodyStatements.push(ts.createVariableStatement(undefined, [resultVariable]));
        if (!this.context.compilationContext.compilerOptions.disableHeapNukeOnExit) {
            // speedyJsGc();
            bodyStatements.push(ts.createStatement(ts.createCall(ts.createPropertyAccess(this.loadWasmFunctionIdentifier, "gc"), [], [])));
        }
        bodyStatements.push(ts.createReturn(resultIdentifier));
        // function (instance) { ... }
        var instanceParameter = ts.createParameter(undefined, undefined, undefined, instanceIdentifier);
        var loadHandlerBody = ts.createBlock(bodyStatements);
        var loadedHandler = ts.createFunctionExpression(undefined, undefined, "instanceLoaded", undefined, [instanceParameter], undefined, loadHandlerBody);
        // loadWasmModule().then(instance => );
        var loadInstance = ts.createCall(this.loadWasmFunctionIdentifier, [], []);
        var instanceLoaded = ts.createCall(ts.createPropertyAccess(loadInstance, "then"), [], [loadedHandler]);
        return ts.updateFunctionDeclaration(functionDeclaration, functionDeclaration.decorators, (functionDeclaration.modifiers || []).filter(function (modifier) { return modifier.kind !== ts.SyntaxKind.AsyncKeyword; }), functionDeclaration.asteriskToken, functionDeclaration.name, functionDeclaration.typeParameters, functionDeclaration.parameters, functionDeclaration.type, ts.createBlock([ts.createReturn(instanceLoaded)]));
    };
    PerFileCodeGeneratorSourceFileRewriter.prototype.rewriteSourceFile = function (sourceFile, requestEmitHelper) {
        assert(this.wasmUrl, "No wasm fetch expression set but requested to transform the source file");
        if (!this.loadWasmFunctionIdentifier) {
            return sourceFile;
        }
        requestEmitHelper(new per_file_wasm_loader_emit_helper_1.PerFileWasmLoaderEmitHelper());
        var compilerOptions = this.context.compilationContext.compilerOptions;
        var options = toLiteral({
            totalStack: compilerOptions.totalStack,
            initialMemory: compilerOptions.initialMemory,
            globalBase: compilerOptions.globalBase,
            staticBump: this.wastMetaData.staticBump || 0,
            exposeGc: compilerOptions.exportGc || compilerOptions.exposeGc
        });
        var initializer = ts.createCall(ts.createIdentifier(per_file_wasm_loader_emit_helper_1.MODULE_LOADER_FACTORY_NAME), [], [
            ts.createLiteral(this.wasmUrl),
            options
        ]);
        var statements = sourceFile.statements;
        var statementsToInsert = [];
        // var loadWasmModule = _moduleLoader(...);
        var loaderDeclaration = ts.createVariableDeclarationList([ts.createVariableDeclaration(this.loadWasmFunctionIdentifier, undefined, initializer)]);
        statementsToInsert.push(ts.createVariableStatement([], loaderDeclaration));
        // export let speedyJsGc = loadWasmModule_1.gc; or without export if only expose
        if (compilerOptions.exposeGc || compilerOptions.exportGc) {
            var speedyJsGcVariable = ts.createVariableDeclaration("speedyJsGc", undefined, ts.createPropertyAccess(this.loadWasmFunctionIdentifier, "gc"));
            var speedyJsGcDeclaration = ts.createVariableDeclarationList([speedyJsGcVariable], ts.NodeFlags.Const);
            var modifiers = compilerOptions.exportGc ? [ts.createToken(ts.SyntaxKind.ExportKeyword)] : [];
            statementsToInsert.push(ts.createVariableStatement(modifiers, speedyJsGcDeclaration));
        }
        return ts.updateSourceFileNode(sourceFile, statementsToInsert.concat(statements));
    };
    PerFileCodeGeneratorSourceFileRewriter.prototype.castToJs = function (returnValue, type, loadWasmFunctionIdentifier, typesIdentifier) {
        // wasm code returns 1 for true and zero for false. Cast it to a boolean
        if (type.flags & ts.TypeFlags.BooleanLike) {
            return ts.createBinary(returnValue, ts.SyntaxKind.EqualsEqualsEqualsToken, ts.createNumericLiteral("1"));
        }
        if (type.flags & ts.TypeFlags.Object || types_1.isMaybeObjectType(type)) {
            return ts.createCall(ts.createPropertyAccess(loadWasmFunctionIdentifier, "toJSObject"), undefined, [
                returnValue,
                ts.createLiteral(serializedTypeName(type, this.context.typeChecker)),
                typesIdentifier
            ]);
        }
        return returnValue;
    };
    PerFileCodeGeneratorSourceFileRewriter.prototype.castToWasm = function (parameterDeclaration, loadWasmFunctionIdentifier, typesIdentifier, argumentObjects) {
        var parameterType = this.context.typeChecker.getTypeAtLocation(parameterDeclaration);
        var symbol = this.context.typeChecker.getSymbolAtLocation(parameterDeclaration.name);
        if (parameterType.flags & ts.TypeFlags.Object || types_1.isMaybeObjectType(parameterType)) {
            return ts.createCall(ts.createPropertyAccess(loadWasmFunctionIdentifier, "toWASM"), undefined, [
                ts.createIdentifier(symbol.name),
                ts.createLiteral(serializedTypeName(parameterType, this.context.typeChecker)),
                typesIdentifier,
                argumentObjects
            ]);
        }
        return ts.createIdentifier(symbol.name);
    };
    PerFileCodeGeneratorSourceFileRewriter.prototype.serializeArgumentAndReturnTypes = function (argumentTypes, returnType) {
        var _this = this;
        var types = {};
        var typesToProcess = Array.from(new Set(argumentTypes.concat([returnType])));
        while (typesToProcess.length > 0) {
            var type = typesToProcess.pop();
            var name = serializedTypeName(type, this.context.typeChecker);
            if (name in types) {
                continue;
            }
            var primitive = !!(type.flags & ts.TypeFlags.BooleanLike || type.flags & ts.TypeFlags.IntLike || type.flags & ts.TypeFlags.NumberLike);
            var fields = [];
            var typeArguments = [];
            var constructor = void 0;
            if (types_1.isMaybeObjectType(type)) {
                type = type.getNonNullableType();
            }
            if (type.flags & ts.TypeFlags.Object) {
                var objectType = type;
                var classReference = this.context.resolveClass(type);
                assert(classReference, "Class Reference for argument or return type " + this.context.typeChecker.typeToString(type) + " not found.");
                fields = classReference.getFields(objectType, this.context).map(function (field) {
                    typesToProcess.push(field.type);
                    return {
                        name: field.name,
                        type: serializedTypeName(field.type, _this.context.typeChecker)
                    };
                });
                if (objectType.objectFlags & ts.ObjectFlags.Reference && objectType.typeArguments) {
                    var typeReference = objectType;
                    typesToProcess.push.apply(typesToProcess, typeReference.typeArguments);
                    typeArguments.push.apply(typeArguments, typeReference.typeArguments.map(function (typeArgument) { return serializedTypeName(typeArgument, _this.context.typeChecker); }));
                }
                constructor = ts.createIdentifier(type.getSymbol().getName());
            }
            types[name] = {
                primitive: primitive,
                fields: fields,
                constructor: constructor,
                typeArguments: typeArguments
            };
        }
        return toLiteral(types);
    };
    return PerFileCodeGeneratorSourceFileRewriter;
}());
exports.PerFileCodeGeneratorSourceFileRewriter = PerFileCodeGeneratorSourceFileRewriter;
function serializedTypeName(type, typeChecker) {
    if (types_1.isMaybeObjectType(type)) {
        type = type.getNonNullableType();
    }
    if (type.flags & ts.TypeFlags.BooleanLike) {
        return "i1";
    }
    else if (type.flags & ts.TypeFlags.IntLike) {
        return "i32";
    }
    else if (type.flags & ts.TypeFlags.NumberLike) {
        return "double";
    }
    else if (type.flags & ts.TypeFlags.Object) {
        var objectType = type;
        var className = typeChecker.getFullyQualifiedName(type.getSymbol());
        if (objectType.objectFlags & ts.ObjectFlags.Reference && objectType.typeArguments) {
            var typeArguments = objectType.typeArguments.map(function (typeArgument) { return serializedTypeName(typeArgument, typeChecker); });
            className += "<" + typeArguments + ">";
        }
        return className;
    }
    else {
        throw new Error("Unsupported type " + typeChecker.typeToString(type));
    }
}
function toLiteral(value) {
    if (Array.isArray(value)) {
        return ts.createArrayLiteral(value.map(function (element) { return toLiteral(element); }));
    }
    else if (typeof value === "object") {
        if (value.kind && value.flags) {
            return value;
        }
        return ts.createObjectLiteral(Object.keys(value).map(function (property) {
            var propertyValue = value[property];
            return ts.createPropertyAssignment(ts.createLiteral(property), toLiteral(propertyValue));
        }));
    }
    else if (typeof value === "undefined") {
        return ts.createIdentifier("undefined");
    }
    else if (typeof value === "number" || typeof value === "boolean" || typeof (value) === "string") {
        return ts.createLiteral(value);
    }
    throw new Error("Unsupported value of type " + typeof (value));
}
//# sourceMappingURL=per-file-code-generator-source-file-rewriter.js.map