"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var debug = require("debug");
var fs = require("fs");
var llvm = require("llvm-node");
var path = require("path");
var ts = require("typescript");
var binaryen_opt_1 = require("../../external-tools/binaryen-opt");
var binaryen_s2wasm_1 = require("../../external-tools/binaryen-s2wasm");
var binaryen_wasm_as_1 = require("../../external-tools/binaryen-wasm-as");
var llvm_link_1 = require("../../external-tools/llvm-link");
var llvm_llc_1 = require("../../external-tools/llvm-llc");
var llvm_opt_1 = require("../../external-tools/llvm-opt");
var build_directory_1 = require("../build-directory");
var default_code_generation_context_factory_1 = require("../default-code-generation-context-factory");
var function_builder_1 = require("../util/function-builder");
var resolved_function_1 = require("../value/resolved-function");
var llvm_emit_source_file_rewriter_1 = require("./llvm-emit-source-file-rewriter");
var per_file_code_generator_source_file_rewriter_1 = require("./per-file-code-generator-source-file-rewriter");
var WASM_TRIPLE = "wasm32-unknown-unknown";
var LOG = debug("code-generation/per-file-code-generator");
/**
 * Code Generator that creates a dedicate WASM Module for each source file
 */
var PerFileCodeGenerator = (function () {
    function PerFileCodeGenerator(context, codeGenerationContextFactory) {
        if (codeGenerationContextFactory === void 0) { codeGenerationContextFactory = new default_code_generation_context_factory_1.DefaultCodeGenerationContextFactory(); }
        this.context = context;
        this.codeGenerationContextFactory = codeGenerationContextFactory;
        this.sourceFileStates = new Map();
    }
    PerFileCodeGenerator.prototype.beginSourceFile = function (sourceFile, compilationContext, requestEmitHelper) {
        var context = this.createContext(sourceFile, compilationContext);
        var emitsWasm = !compilationContext.compilerOptions.emitLLVM;
        this.sourceFileStates.set(sourceFile.fileName, {
            context: context,
            requestEmitHelper: requestEmitHelper,
            sourceFileRewriter: emitsWasm ? new per_file_code_generator_source_file_rewriter_1.PerFileCodeGeneratorSourceFileRewriter(context) : new llvm_emit_source_file_rewriter_1.NoopSourceFileRewriter()
        });
    };
    PerFileCodeGenerator.prototype.generateEntryFunction = function (functionDeclaration) {
        var state = this.getSourceFileState(functionDeclaration);
        var context = state.context;
        // function is async, therefore, return type is a promise of T. We compile it down to a function just returning T
        var signature = context.typeChecker.getSignatureFromDeclaration(functionDeclaration);
        var resolvedFunction = resolved_function_1.createResolvedFunctionFromSignature(signature, context.compilationContext);
        var mangledName = "_" + resolvedFunction.functionName;
        var builder = function_builder_1.FunctionBuilder.create(resolvedFunction, context)
            .name(mangledName)
            .externalLinkage();
        builder.define(resolvedFunction.definition);
        context.addEntryFunction(mangledName);
        return state.sourceFileRewriter.rewriteEntryFunction(mangledName, functionDeclaration, state.requestEmitHelper);
    };
    PerFileCodeGenerator.prototype.completeSourceFile = function (sourceFile) {
        var state = this.getSourceFileState(sourceFile);
        var context = state.context;
        if (context.module.empty || context.getEntryFunctionNames().length === 0) {
            return sourceFile;
        }
        if (context.requiresGc) {
            context.module.getOrInsertFunction("speedyJsGc", llvm.FunctionType.get(llvm.Type.getVoidTy(context.llvmContext), [], false));
        }
        LOG("Emit module for source file " + sourceFile.fileName + ".");
        llvm.verifyModule(context.module);
        this.writeModule(sourceFile, state);
        LOG("Module for source file " + sourceFile.fileName + " emitted.");
        return state.sourceFileRewriter.rewriteSourceFile(sourceFile, state.requestEmitHelper);
    };
    PerFileCodeGenerator.prototype.completeCompilation = function () {
        this.sourceFileStates.clear();
    };
    PerFileCodeGenerator.prototype.writeModule = function (sourceFile, state) {
        var context = state.context;
        var buildDirectory = build_directory_1.BuildDirectory.createTempBuildDirectory();
        var plainFileName = path.basename(sourceFile.fileName.replace(".ts", ""));
        if (context.compilationContext.compilerOptions.emitLLVM) {
            var llc_1 = context.module.print();
            context.compilationContext.compilerHost.writeFile(getOutputFileName(sourceFile, context, ".ll"), llc_1, false);
        }
        else {
            var transforms = PerFileCodeGenerator.createTransformationChain(context);
            var transformationContext_1 = {
                sourceFile: sourceFile,
                codeGenerationContext: context,
                buildDirectory: buildDirectory,
                plainFileName: plainFileName,
                sourceFileRewriter: state.sourceFileRewriter
            };
            var biteCodeFileName = buildDirectory.getTempFileName(plainFileName + ".bc");
            llvm.writeBitcodeToFile(context.module, biteCodeFileName);
            transforms.reduce(function (inputFileName, transformation) { return transformation.transform(inputFileName, transformationContext_1); }, biteCodeFileName);
        }
        buildDirectory.remove();
    };
    PerFileCodeGenerator.createTransformationChain = function (context) {
        var optimizationLevel = context.compilationContext.compilerOptions.optimizationLevel;
        var transforms = [
            LinkTransformationStep.createRuntimeLinking()
        ];
        if (optimizationLevel !== "0") {
            transforms.push(new OptimizationTransformationStep());
        }
        transforms.push(LinkTransformationStep.createSharedLibsLinking(), new LinkTimeOptimizationTransformationStep());
        if (context.compilationContext.compilerOptions.saveBc) {
            transforms.push(new CopyFileToOutputDirectoryTransformationStep(".bc", true));
        }
        transforms.push(new LLCTransformationStep(), new S2WasmTransformationStep());
        if (context.compilationContext.compilerOptions.binaryenOpt && optimizationLevel !== "0") {
            transforms.push(new BinaryenOptTransformationStep());
        }
        if (context.compilationContext.compilerOptions.saveWast) {
            transforms.push(new CopyFileToOutputDirectoryTransformationStep(".wast", true));
        }
        transforms.push(new WasmToAsTransformationStep());
        return transforms;
    };
    PerFileCodeGenerator.prototype.getSourceFileState = function (node) {
        var sourceFile = node.getSourceFile();
        var state = this.sourceFileStates.get(sourceFile.fileName);
        // tslint:disable-next-line:max-line-length
        assert(state, "First call beginSourceFile before calling any other functions on the code generator (state missing for source file " + sourceFile.fileName + ").");
        return state;
    };
    PerFileCodeGenerator.prototype.createContext = function (sourceFile, compilationContext) {
        var relativePath = path.relative(compilationContext.rootDir, sourceFile.fileName);
        var module = new llvm.Module(sourceFile.moduleName || relativePath, this.context);
        module.sourceFileName = relativePath;
        var target = llvm.TargetRegistry.lookupTarget(WASM_TRIPLE);
        var targetMachine = target.createTargetMachine(WASM_TRIPLE, "generic");
        module.dataLayout = targetMachine.createDataLayout();
        module.targetTriple = WASM_TRIPLE;
        return this.codeGenerationContextFactory.createContext(compilationContext, module);
    };
    return PerFileCodeGenerator;
}());
exports.PerFileCodeGenerator = PerFileCodeGenerator;
var OptimizationTransformationStep = (function () {
    function OptimizationTransformationStep() {
    }
    OptimizationTransformationStep.prototype.transform = function (inputFileName, _a) {
        var plainFileName = _a.plainFileName, buildDirectory = _a.buildDirectory, codeGenerationContext = _a.codeGenerationContext;
        var optimizedFileName = buildDirectory.getTempFileName(plainFileName + "-opt.bc");
        var optimizationLevel = codeGenerationContext.compilationContext.compilerOptions.optimizationLevel;
        return llvm_opt_1.optimize(inputFileName, optimizedFileName, optimizationLevel);
    };
    return OptimizationTransformationStep;
}());
var LinkTransformationStep = (function () {
    function LinkTransformationStep(runtime) {
        this.runtime = runtime;
    }
    LinkTransformationStep.createRuntimeLinking = function () {
        return new LinkTransformationStep(true);
    };
    LinkTransformationStep.createSharedLibsLinking = function () {
        return new LinkTransformationStep(false);
    };
    LinkTransformationStep.prototype.transform = function (inputFileName, _a) {
        var plainFileName = _a.plainFileName, buildDirectory = _a.buildDirectory, codeGenerationContext = _a.codeGenerationContext;
        var llvmLinker = new llvm_link_1.LLVMLink(buildDirectory);
        var entryFunctions = codeGenerationContext.getEntryFunctionNames();
        llvmLinker.addByteCodeFile(inputFileName);
        if (this.runtime) {
            llvmLinker.addRuntime(codeGenerationContext.compilationContext.compilerOptions.unsafe);
        }
        else {
            llvmLinker.addSharedLibs();
        }
        return llvmLinker.link(buildDirectory.getTempFileName(plainFileName + "-linked.o"), entryFunctions);
    };
    return LinkTransformationStep;
}());
var LinkTimeOptimizationTransformationStep = (function () {
    function LinkTimeOptimizationTransformationStep() {
    }
    LinkTimeOptimizationTransformationStep.prototype.transform = function (inputFileName, _a) {
        var plainFileName = _a.plainFileName, buildDirectory = _a.buildDirectory, codeGenerationContext = _a.codeGenerationContext;
        var optimizedFileName = buildDirectory.getTempFileName(plainFileName + "-lopt.bc");
        var entryFunctionNames = codeGenerationContext.getEntryFunctionNames();
        return llvm_opt_1.optimizeLinked(inputFileName, entryFunctionNames, optimizedFileName, codeGenerationContext.compilationContext.compilerOptions.optimizationLevel);
    };
    return LinkTimeOptimizationTransformationStep;
}());
var CopyFileToOutputDirectoryTransformationStep = (function () {
    function CopyFileToOutputDirectoryTransformationStep(extension, binary) {
        if (binary === void 0) { binary = false; }
        this.extension = extension;
        this.binary = binary;
    }
    CopyFileToOutputDirectoryTransformationStep.prototype.transform = function (inputFileName, _a) {
        var plainFileName = _a.plainFileName, buildDirectory = _a.buildDirectory, sourceFile = _a.sourceFile, codeGenerationContext = _a.codeGenerationContext;
        var outputFileName = getOutputFileName(sourceFile, codeGenerationContext, this.extension);
        if (this.binary) {
            // ts writeFile seems not to support binary output, therefore use fs directly
            fs.writeFileSync(outputFileName, fs.readFileSync(inputFileName));
        }
        else {
            // has the advantage that it can be intercepted by the in memory compiler
            codeGenerationContext.compilationContext.compilerHost.writeFile(outputFileName, ts.sys.readFile(inputFileName), false);
        }
        return inputFileName;
    };
    return CopyFileToOutputDirectoryTransformationStep;
}());
var LLCTransformationStep = (function () {
    function LLCTransformationStep() {
    }
    LLCTransformationStep.prototype.transform = function (inputFileName, _a) {
        var plainFileName = _a.plainFileName, buildDirectory = _a.buildDirectory;
        return llvm_llc_1.llc(inputFileName, buildDirectory.getTempFileName(plainFileName + ".s"));
    };
    return LLCTransformationStep;
}());
var S2WasmTransformationStep = (function () {
    function S2WasmTransformationStep() {
    }
    S2WasmTransformationStep.prototype.transform = function (inputFileName, _a) {
        var buildDirectory = _a.buildDirectory, plainFileName = _a.plainFileName, codeGenerationContext = _a.codeGenerationContext, sourceFileRewriter = _a.sourceFileRewriter;
        var compilerOptions = codeGenerationContext.compilationContext.compilerOptions;
        var wastFileName = binaryen_s2wasm_1.s2wasm(inputFileName, buildDirectory.getTempFileName(plainFileName + ".wast"), compilerOptions);
        var glue = S2WasmTransformationStep.getGlueMetadata(wastFileName);
        sourceFileRewriter.setWastMetaData(glue);
        return wastFileName;
    };
    // glue code https://github.com/kripken/emscripten/blob/5387c1e5f1f55b69c41b94cf044062c59e052f0b/emscripten.py#L1415
    S2WasmTransformationStep.getGlueMetadata = function (wastFileName) {
        var wastContent = ts.sys.readFile(wastFileName);
        var parts = wastContent.split("\n;; METADATA:");
        var metaDataRaw = parts[1];
        return JSON.parse(metaDataRaw);
    };
    return S2WasmTransformationStep;
}());
var BinaryenOptTransformationStep = (function () {
    function BinaryenOptTransformationStep() {
    }
    BinaryenOptTransformationStep.prototype.transform = function (inputFileName, _a) {
        var buildDirectory = _a.buildDirectory, plainFileName = _a.plainFileName, codeGenerationContext = _a.codeGenerationContext;
        var outputFile = buildDirectory.getTempFileName(plainFileName + "-opt.wast");
        return binaryen_opt_1.wasmOpt(inputFileName, outputFile);
    };
    return BinaryenOptTransformationStep;
}());
var WasmToAsTransformationStep = (function () {
    function WasmToAsTransformationStep() {
    }
    WasmToAsTransformationStep.prototype.transform = function (inputFileName, _a) {
        var buildDirectory = _a.buildDirectory, plainFileName = _a.plainFileName, sourceFileRewriter = _a.sourceFileRewriter, sourceFile = _a.sourceFile, codeGenerationContext = _a.codeGenerationContext;
        var wasmFileName = buildDirectory.getTempFileName(plainFileName + ".wasm");
        binaryen_wasm_as_1.wasmAs(inputFileName, wasmFileName);
        var wasmFileWriter = codeGenerationContext.compilationContext.compilerOptions.wasmFileWriter;
        var finalWasmFileName = getOutputFileName(sourceFile, codeGenerationContext);
        var wasmFetchExpression = wasmFileWriter.writeWasmFile(finalWasmFileName, fs.readFileSync(wasmFileName), codeGenerationContext);
        sourceFileRewriter.setWasmUrl(wasmFetchExpression);
        return wasmFileName;
    };
    return WasmToAsTransformationStep;
}());
function getOutputFileName(sourceFile, context, fileExtension) {
    if (fileExtension === void 0) { fileExtension = ".wasm"; }
    var withNewExtension = sourceFile.fileName.replace(".ts", fileExtension);
    if (context.compilationContext.compilerOptions.outDir) {
        var relativeName = path.relative(context.compilationContext.rootDir, withNewExtension);
        return path.join(context.compilationContext.compilerOptions.outDir, relativeName);
    }
    return withNewExtension;
}
//# sourceMappingURL=per-file-code-generator.js.map