"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
/**
 * The lexical scope
 */
var Scope = (function () {
    function Scope(parent, fn) {
        this.parent = parent;
        this.fn = fn;
        this.variables = new Map();
        this.labels = new Map();
        this.children = [];
        this.functions = parent ? parent.functions : new Map();
        this.classes = parent ? parent.classes : new Map();
    }
    Object.defineProperty(Scope.prototype, "returnAllocation", {
        /**
         * Stores the function result. Only present in a function that is non void
         */
        get: function () {
            return this.returnAlloca || (this.parent ? this.parent.returnAllocation : undefined);
        },
        set: function (allocation) {
            this.returnAlloca = allocation;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Returns the block that is the target of a continue statement
     * @param label the name of the label or undefined for the default continue target
     * @return the basic block that is the target of the continue statement or undefined
     */
    Scope.prototype.getContinueBlock = function (label) {
        return this.getLabel(typeof (label) === "undefined" ? "continue" : label + "-continue");
    };
    /**
     * Sets the target block for the continue statement
     * @param block the target block
     * @param label the name of the label or undefined for the continue statement without a label
     */
    Scope.prototype.setContinueBlock = function (block, label) {
        this.labels.set(typeof (label) === "undefined" ? "continue" : label + "-continue", block);
    };
    /**
     * Returns the block that is the target of a break statement
     * @param label the name of the label or undefined for the default break target
     * @return the basic block that is the target of the break statement or undefined
     */
    Scope.prototype.getBreakBlock = function (label) {
        return this.getLabel(typeof (label) === "undefined" ? "break" : label + "-break");
    };
    /**
     * Sets the target block for the break statement
     * @param block the target block
     * @param label the name of the label or undefined for the break statement without a label
     */
    Scope.prototype.setBreakBlock = function (block, label) {
        this.labels.set(typeof (label) === "undefined" ? "break" : label + "-break", block);
    };
    Scope.prototype.getLabel = function (label) {
        return this.labels.get(label) || (this.parent ? this.parent.getLabel(label) : undefined);
    };
    Object.defineProperty(Scope.prototype, "returnBlock", {
        /**
         * Block that is the terminator of a function. All return statements need to jump directly to this return block
         * to ensure that no succeeding statements are executed. Only present inside of a function (also void function, as these might return void);
         */
        get: function () {
            return this.getLabel("return") || (this.parent ? this.parent.returnBlock : undefined);
        },
        set: function (returnBlock) {
            if (returnBlock) {
                this.labels.set("return", returnBlock);
            }
            else {
                this.labels.delete("return");
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Scope.prototype, "enclosingFunction", {
        /**
         * Returns a reference to the function in which this scope is defined
         * @return the function
         */
        get: function () {
            var result = this.fn || (this.parent ? this.parent.enclosingFunction : undefined);
            assert(result, "Code generation always needs to be inside of a function");
            return result;
        },
        enumerable: true,
        configurable: true
    });
    Scope.prototype.addVariable = function (symbol, value) {
        assert(symbol, "symbol is undefined");
        assert(value, "value is undefined");
        assert(!this.variables.has(symbol), "Variable " + symbol.name + " is already defined in scope");
        this.variables.set(symbol, value);
    };
    Scope.prototype.getVariable = function (symbol) {
        assert(symbol, "symbol is undefined");
        var variable = this.variables.get(symbol);
        if (!variable && this.parent) {
            return this.parent.getVariable(symbol);
        }
        assert(variable, "variable " + symbol.name + " is not defined in scope");
        return variable;
    };
    Scope.prototype.getVariables = function () {
        return Array.from(this.variables.keys());
    };
    Scope.prototype.getAllVariables = function () {
        var childVariables = this.children.map(function (child) { return child.getAllVariables(); });
        return Array.prototype.concat.apply(this.getVariables(), childVariables);
    };
    Scope.prototype.hasVariable = function (symbol) {
        return this.variables.has(symbol) || (!!this.parent && this.parent.hasVariable(symbol));
    };
    Scope.prototype.addFunction = function (symbol, fn) {
        assert(fn, "function is undefined");
        assert(symbol, "symbol is undefined");
        assert(!this.functions.has(symbol), "function " + symbol.name + " is already defined in scope");
        this.functions.set(symbol, fn);
    };
    Scope.prototype.getFunction = function (symbol) {
        assert(symbol, "symbol is undefined");
        var fun = this.functions.get(symbol);
        assert(fun, "function " + symbol.name + " is not defined in scope");
        return fun;
    };
    Scope.prototype.addClass = function (symbol, classReference) {
        assert(symbol, "symbol is undefined");
        assert(classReference, "class reference is undefined");
        assert(!this.classes.has(symbol), "class " + symbol.name + " is already defined");
        this.classes.set(symbol, classReference);
    };
    Scope.prototype.getClass = function (symbol) {
        assert(symbol, "symbol is undefined");
        var cls = this.classes.get(symbol);
        assert(cls, "Class " + symbol.name + " is not defined in scope");
        return cls;
    };
    Scope.prototype.enterChild = function (fn) {
        var child = new Scope(this, fn);
        this.children.push(child);
        return child;
    };
    Scope.prototype.exitChild = function () {
        assert(this.parent, "Cannot leave root scope");
        return this.parent;
    };
    Scope.prototype.hasFunction = function (symbol) {
        return this.functions.has(symbol);
    };
    Scope.prototype.hasClass = function (symbol) {
        return this.classes.has(symbol);
    };
    return Scope;
}());
exports.Scope = Scope;
//# sourceMappingURL=scope.js.map