"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var llvm = require("llvm-node");
var default_name_mangler_1 = require("../default-name-mangler");
var runtime_system_name_mangler_1 = require("../runtime-system-name-mangler");
var object_property_reference_1 = require("../value/object-property-reference");
var types_1 = require("./types");
var ComputedObjectPropertyReferenceBuilder = (function () {
    function ComputedObjectPropertyReferenceBuilder(property, context) {
        this.property = property;
        this.context = context;
        this.runtimeFn = false;
        this.readOnlyFlag = false;
    }
    ComputedObjectPropertyReferenceBuilder.forProperty = function (property, context) {
        return new ComputedObjectPropertyReferenceBuilder(property, context);
    };
    ComputedObjectPropertyReferenceBuilder.prototype.fromRuntime = function () {
        this.runtimeFn = true;
        return this;
    };
    ComputedObjectPropertyReferenceBuilder.prototype.readonly = function (isReadonly) {
        if (isReadonly === void 0) { isReadonly = true; }
        this.readOnlyFlag = isReadonly;
        return this;
    };
    ComputedObjectPropertyReferenceBuilder.prototype.build = function (objectReference) {
        var propertyType = this.context.typeChecker.getTypeAtLocation(this.property);
        var property = this.context.typeChecker.getSymbolAtLocation(this.property);
        var propertyLLVMType = types_1.toLLVMType(propertyType, this.context);
        var thisLLVMType = types_1.toLLVMType(objectReference.type, this.context);
        var getter = this.createGetter(thisLLVMType, propertyLLVMType, objectReference);
        var setter;
        if (!this.readOnlyFlag) {
            setter = this.createSetter(thisLLVMType, propertyLLVMType, objectReference);
        }
        return object_property_reference_1.ObjectPropertyReference.createComputedPropertyReference(propertyType, objectReference, property, getter, setter);
    };
    ComputedObjectPropertyReferenceBuilder.prototype.getNameMangler = function () {
        return this.runtimeFn ? new runtime_system_name_mangler_1.RuntimeSystemNameMangler(this.context.compilationContext) : new default_name_mangler_1.DefaultNameMangler(this.context.compilationContext);
    };
    ComputedObjectPropertyReferenceBuilder.prototype.createGetter = function (thisType, propertyType, objectReference) {
        var getterName = this.getNameMangler().mangleProperty(this.property, false);
        var getter = this.context.module.getFunction(getterName);
        if (!getter) {
            var getterType = llvm.FunctionType.get(propertyType, [thisType], false);
            getter = llvm.Function.create(getterType, llvm.LinkageTypes.ExternalLinkage, getterName, this.context.module);
            getter.addFnAttr(llvm.Attribute.AttrKind.AlwaysInline);
            getter.addFnAttr(llvm.Attribute.AttrKind.ReadOnly);
            getter.addFnAttr(llvm.Attribute.AttrKind.NoUnwind);
            var self = getter.getArguments()[0];
            self.addAttr(llvm.Attribute.AttrKind.ReadOnly);
            self.addAttr(llvm.Attribute.AttrKind.NoCapture);
            self.addDereferenceableAttr(objectReference.getTypeStoreSize(this.context));
        }
        return getter;
    };
    ComputedObjectPropertyReferenceBuilder.prototype.createSetter = function (thisType, propertyType, objectReference) {
        var setterName = this.getNameMangler().mangleProperty(this.property, true);
        var setter = this.context.module.getFunction(setterName);
        if (!setter) {
            var setterType = llvm.FunctionType.get(llvm.Type.getVoidTy(this.context.llvmContext), [thisType, propertyType], false);
            setter = llvm.Function.create(setterType, llvm.LinkageTypes.ExternalLinkage, setterName, this.context.module);
            setter.addFnAttr(llvm.Attribute.AttrKind.AlwaysInline);
            var self = setter.getArguments()[0];
            self.addDereferenceableAttr(objectReference.getTypeStoreSize(this.context));
        }
        return setter;
    };
    return ComputedObjectPropertyReferenceBuilder;
}());
exports.ComputedObjectPropertyReferenceBuilder = ComputedObjectPropertyReferenceBuilder;
//# sourceMappingURL=computed-object-property-reference-builder.js.map