"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var llvm = require("llvm-node");
var ts = require("typescript");
var primitive_1 = require("../value/primitive");
/**
 * Generates a condition statement that takes the whenTrue branch if the condition evaluates to a truthy value and to the whenFalse branch otherwise.
 * The implementation optimizes conditions. E.g. && and || are optimized to not store the value and instead just evaluate
 * if the expression is truthy (the value is not needed)
 * @param condition the condition
 * @param whenTrue the block to take if the condition is true
 * @param whenFalse the block to take if the condition is false
 * @param context the context
 */
function generateCondition(condition, whenTrue, whenFalse, context) {
    if (condition.kind === ts.SyntaxKind.ParenthesizedExpression) {
        return generateCondition(condition.expression, whenTrue, whenFalse, context);
    }
    if (condition.kind === ts.SyntaxKind.BinaryExpression) {
        var binaryExpression = condition;
        // x && y
        if (binaryExpression.operatorToken.kind === ts.SyntaxKind.AmpersandAmpersandToken) {
            var rhsBlock = llvm.BasicBlock.create(context.llvmContext, "land.lhs.true");
            generateCondition(binaryExpression.left, rhsBlock, whenFalse, context);
            context.scope.enclosingFunction.addBasicBlock(rhsBlock);
            context.builder.setInsertionPoint(rhsBlock);
            generateCondition(binaryExpression.right, whenTrue, whenFalse, context);
            return;
        }
        else if (binaryExpression.operatorToken.kind === ts.SyntaxKind.BarBarToken) {
            var rhsBlock = llvm.BasicBlock.create(context.llvmContext, "lor.lhs.false");
            generateCondition(binaryExpression.left, whenTrue, rhsBlock, context);
            context.scope.enclosingFunction.addBasicBlock(rhsBlock);
            context.builder.setInsertionPoint(rhsBlock);
            generateCondition(binaryExpression.right, whenTrue, whenFalse, context);
            return;
        }
    }
    else if (condition.kind === ts.SyntaxKind.PrefixUnaryExpression) {
        var unaryExpression = condition;
        if (unaryExpression.operator === ts.SyntaxKind.ExclamationToken) {
            generateCondition(unaryExpression.operand, whenFalse, whenTrue, context);
            return;
        }
    }
    // Not specially optimized case
    var conditionValue = context.generateValue(condition);
    var conditionBoolValue = primitive_1.Primitive.toBoolean(conditionValue, context.typeChecker.getTypeAtLocation(condition), context);
    context.builder.createCondBr(conditionBoolValue, whenTrue, whenFalse);
}
exports.generateCondition = generateCondition;
//# sourceMappingURL=conditions.js.map