"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var llvm = require("llvm-node");
var resolved_function_reference_1 = require("../value/resolved-function-reference");
var unresolved_function_reference_1 = require("../value/unresolved-function-reference");
var function_declaration_builder_1 = require("./function-declaration-builder");
var function_definition_builder_1 = require("./function-definition-builder");
/**
 * Builder for declaring and defining llvm functions
 */
var FunctionBuilder = (function () {
    function FunctionBuilder(resolvedFunction, context) {
        this.resolvedFunction = resolvedFunction;
        this.context = context;
        this.declarationBuilder = function_declaration_builder_1.FunctionDeclarationBuilder.forResolvedFunction(resolvedFunction, context);
        this.functionName = resolvedFunction.functionName;
    }
    /**
     * Creates a builder for declaring the given resolved function
     * @param resolvedFunction the resolved function to declare
     * @param context the code generation context
     * @return {FunctionDeclarationBuilder} the builder instance
     */
    FunctionBuilder.create = function (resolvedFunction, context) {
        return new FunctionBuilder(resolvedFunction, context);
    };
    /**
     * Sets the function as externally linked
     * @return this for a fluent api
     */
    FunctionBuilder.prototype.externalLinkage = function () {
        return this.linkage(llvm.LinkageTypes.ExternalLinkage);
    };
    /**
     * Sets the function as internally linked
     * @return this for a fluent api
     */
    FunctionBuilder.prototype.internalLinkage = function () {
        return this.linkage(llvm.LinkageTypes.InternalLinkage);
    };
    /**
     * Changes the linkage of the function to the value passed
     * @param linkage the desired linkage
     * @return this for a fluent api
     */
    FunctionBuilder.prototype.linkage = function (linkage) {
        this.declarationBuilder.linkage(linkage);
        return this;
    };
    /**
     * Changes the name of the function
     * @param name the name of the function
     * @return {FunctionBuilder} this for a fluent api
     */
    FunctionBuilder.prototype.name = function (name) {
        this.functionName = name;
        return this;
    };
    /**
     * Builds / Generates the llvm.Function for the given function declaration
     * @param declaration the function declaration
     * @return {FunctionReference} the reference for the generated function
     */
    FunctionBuilder.prototype.define = function (declaration) {
        var _this = this;
        assert(declaration.body, "Cannot transform function declaration without a body");
        var fun = this.declarationBuilder.name(this.functionName).declare();
        var functionReference = resolved_function_reference_1.ResolvedFunctionReference.create(fun, this.resolvedFunction);
        function_definition_builder_1.FunctionDefinitionBuilder.create(fun, this.resolvedFunction, this.context).define();
        if (this.resolvedFunction.symbol) {
            // Function is overloaded, determine overload when function is dereferenced
            var declarations = this.resolvedFunction.symbol.declarations;
            if (declarations && declarations.length > 1) {
                var signatures = declarations.map(function (functionDeclaration) { return _this.context.typeChecker.getSignatureFromDeclaration(functionDeclaration); });
                var unresolvedFunction = unresolved_function_reference_1.UnresolvedFunctionReference.createFunction(signatures, this.context, this.resolvedFunction.classType);
                this.context.scope.addFunction(this.resolvedFunction.symbol, unresolvedFunction);
            }
            else {
                // optimization for not overloaded functions
                this.context.scope.addFunction(this.resolvedFunction.symbol, functionReference);
            }
        }
        return functionReference;
    };
    return FunctionBuilder;
}());
exports.FunctionBuilder = FunctionBuilder;
//# sourceMappingURL=function-builder.js.map