"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var llvm = require("llvm-node");
var types_1 = require("./types");
/**
 * Builder for declaring llvm functions
 */
var FunctionDeclarationBuilder = (function () {
    function FunctionDeclarationBuilder(functionName, parameters, returnType, context) {
        this.functionName = functionName;
        this.parameters = parameters;
        this.returnType = returnType;
        this.context = context;
        this.linkageType = llvm.LinkageTypes.InternalLinkage;
        this.attributes = [];
    }
    /**
     * Creates a builder for the given resolved function
     * @param resolvedFunction the resolvedFunction
     * @param context the code generation context
     * @param returnType in case the return type differs from the one specified on the resolved function
     * @return {FunctionDeclarationBuilder} the created builder instance to declare a function with the given signature
     */
    FunctionDeclarationBuilder.forResolvedFunction = function (resolvedFunction, context, returnType) {
        return this.create(resolvedFunction.functionName, resolvedFunction.parameters, returnType || resolvedFunction.returnType, context);
    };
    /**
     * Creates a builder to declare a function
     * @param name the name of the function
     * @param parameters the parameters of the function and their type
     * @param returnType the return type of the function
     * @param context the context
     * @return the builder
     */
    FunctionDeclarationBuilder.create = function (name, parameters, returnType, context) {
        return new FunctionDeclarationBuilder(name, parameters, returnType, context);
    };
    /**
     * Changes the name of the function
     * @param name the new name
     * @return this for a fluent api
     */
    FunctionDeclarationBuilder.prototype.name = function (name) {
        this.functionName = name;
        return this;
    };
    /**
     * Sets the function as externally linked
     * @return this for a fluent api
     */
    FunctionDeclarationBuilder.prototype.externalLinkage = function () {
        return this.linkage(llvm.LinkageTypes.ExternalLinkage);
    };
    /**
     * Sets the function as link once ODR linked
     * @return this for a fluent api
     */
    FunctionDeclarationBuilder.prototype.linkOnceOdrLinkage = function () {
        return this.linkage(llvm.LinkageTypes.LinkOnceODRLinkage);
    };
    /**
     * Adds the given function attribute
     * @param attribute the attribute to add
     * @return {FunctionDeclarationBuilder}
     */
    FunctionDeclarationBuilder.prototype.withAttribute = function (attribute) {
        this.attributes.push(attribute);
        return this;
    };
    /**
     * Changes the linkage of the function to the value passed
     * @param linkage the desired linkage
     * @return this for a fluent api
     */
    FunctionDeclarationBuilder.prototype.linkage = function (linkage) {
        this.linkageType = linkage;
        return this;
    };
    /**
     * Generates only the header of the function and registers the function in the scope (important to be able to handle recursive functions)
     * @return this for a fluent api
     */
    FunctionDeclarationBuilder.prototype.declare = function () {
        var llvmReturnType = types_1.toLLVMType(this.returnType, this.context);
        var parameters = [];
        for (var _i = 0, _a = this.parameters; _i < _a.length; _i++) {
            var parameter = _a[_i];
            parameters.push(types_1.toLLVMType(parameter.type, this.context));
        }
        var functionType = llvm.FunctionType.get(llvmReturnType, parameters, false);
        var fn = llvm.Function.create(functionType, this.linkageType, this.functionName, this.context.module);
        for (var _b = 0, _c = this.attributes; _b < _c.length; _b++) {
            var attribute = _c[_b];
            fn.addFnAttr(attribute);
        }
        return fn;
    };
    FunctionDeclarationBuilder.prototype.declareIfNotExisting = function () {
        var existing = this.context.module.getFunction(this.functionName);
        if (existing) {
            return existing;
        }
        return this.declare();
    };
    return FunctionDeclarationBuilder;
}());
exports.FunctionDeclarationBuilder = FunctionDeclarationBuilder;
//# sourceMappingURL=function-declaration-builder.js.map