"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var llvm = require("llvm-node");
var ts = require("typescript");
var allocation_1 = require("../value/allocation");
var array_class_reference_1 = require("../value/array-class-reference");
var FunctionDefinitionBuilder = (function () {
    function FunctionDefinitionBuilder(fn, resolvedFunction, context) {
        this.fn = fn;
        this.resolvedFunction = resolvedFunction;
        this.context = context;
        // tslint:disable-next-line:variable-name
        this._returnValue = undefined;
    }
    FunctionDefinitionBuilder.create = function (fn, resolvedFunction, context) {
        assert(resolvedFunction.declaration, "Resolved function misses declaration and, therefore, cannot be defined.");
        assert(resolvedFunction.definition, "Resolved function misses definition and, therefore, cannot be defined.");
        return new FunctionDefinitionBuilder(fn, resolvedFunction, context);
    };
    Object.defineProperty(FunctionDefinitionBuilder.prototype, "definition", {
        get: function () {
            return this.resolvedFunction.definition;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(FunctionDefinitionBuilder.prototype, "declaration", {
        get: function () {
            return this.resolvedFunction.declaration;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Sets the value to return at the end of the function
     * @return {FunctionDefinitionBuilder} this for a fluent api
     */
    FunctionDefinitionBuilder.prototype.returnValue = function (returnValue) {
        this._returnValue = returnValue;
        return this;
    };
    /**
     * Sets the address of the this object in case it is not passed as argument
     * @param thisObject the address of the this object
     * @return {FunctionDefinitionBuilder}
     */
    FunctionDefinitionBuilder.prototype.self = function (thisObject) {
        this._self = thisObject;
        return this;
    };
    /**
     * Builds / Generates the llvm.Function for the function definition of the resolved function
     */
    FunctionDefinitionBuilder.prototype.define = function () {
        this.context.enterChildScope(this.fn);
        var entryBlock = this.fn.getEntryBlock() || llvm.BasicBlock.create(this.context.llvmContext, "entry", this.fn);
        var returnBlock = llvm.BasicBlock.create(this.context.llvmContext, "returnBlock");
        this.context.builder.setInsertionPoint(entryBlock);
        this.context.scope.returnBlock = returnBlock;
        if (!(this.resolvedFunction.returnType.flags & ts.TypeFlags.Void) && !this._returnValue) {
            this.context.scope.returnAllocation = allocation_1.Allocation.create(this.resolvedFunction.returnType, this.context, "return");
        }
        this.allocateArguments();
        this.context.generate(this.definition.body);
        this.setBuilderToReturnBlock(returnBlock);
        this.generateReturnStatement();
        this.context.leaveChildScope();
        llvm.verifyFunction(this.fn);
    };
    FunctionDefinitionBuilder.prototype.generateReturnStatement = function () {
        if (this.context.scope.returnAllocation) {
            this.context.builder.createRet(this.context.scope.returnAllocation.generateIR(this.context));
        }
        else if (this._returnValue) {
            this.context.builder.createRet(this._returnValue.generateIR(this.context));
        }
        else {
            this.context.builder.createRetVoid();
        }
    };
    FunctionDefinitionBuilder.prototype.setBuilderToReturnBlock = function (returnBlock) {
        // Current Block is empty, so we can use the current block instead of the return block
        var predecessor = this.context.builder.getInsertBlock();
        if (predecessor.empty) {
            returnBlock.replaceAllUsesWith(predecessor);
            returnBlock.release();
        }
        else if (returnBlock.useEmpty()) {
            returnBlock.release();
        }
        else {
            if (!predecessor.getTerminator()) {
                this.context.builder.createBr(returnBlock); // Fall through to return block
            }
            this.fn.addBasicBlock(returnBlock);
            this.context.builder.setInsertionPoint(returnBlock);
        }
    };
    FunctionDefinitionBuilder.prototype.allocateArguments = function () {
        var args = this.fn.getArguments().slice();
        // The this object is passed as first argument
        if (this.resolvedFunction.classType && this.resolvedFunction.instanceMethod) {
            assert(args.length - 1 === this.resolvedFunction.parameters.length, "The function declaration has no additional argument for the this object");
            var thisAllocation = allocation_1.Allocation.create(this.resolvedFunction.classType, this.context, "this");
            var thisArg = args.shift();
            thisAllocation.generateAssignmentIR(thisArg, this.context);
            thisArg.name = "this";
            this.context.scope.addVariable(this.resolvedFunction.classType.getSymbol(), thisAllocation);
        }
        for (var i = 0; i < this.resolvedFunction.parameters.length; ++i) {
            var parameter = this.resolvedFunction.parameters[i];
            var parameterDefinition = this.definition.parameters[i];
            var declaredParameterSymbol = this.context.typeChecker.getSymbolAtLocation(parameterDefinition.name);
            var allocation = allocation_1.Allocation.create(parameter.type, this.context, parameter.name + ".addr");
            args[i].name = parameter.name;
            if (parameter.variadic) {
                var arrayOfVarArgs = array_class_reference_1.ArrayClassReference.fromCArray(parameter.type, args[i], args[++i], this.context);
                allocation.generateAssignmentIR(arrayOfVarArgs, this.context);
            }
            else {
                allocation.generateAssignmentIR(args[i], this.context);
            }
            this.context.scope.addVariable(declaredParameterSymbol, allocation);
            // a field in a constructor that is marked with private, protected or public. Set the argument value on the field.
            if (this._self && declaredParameterSymbol.flags & ts.SymbolFlags.Property) {
                var fieldOffset = llvm.ConstantInt.get(this.context.llvmContext, this._self.clazz.getFieldOffset(declaredParameterSymbol));
                var fieldIndex = [llvm.ConstantInt.get(this.context.llvmContext, 0), fieldOffset];
                var thiz = this._self.generateIR(this.context);
                var fieldAddress = this.context.builder.createInBoundsGEP(thiz, fieldIndex, "&" + declaredParameterSymbol.name);
                this.context.builder.createAlignedStore(args[i], fieldAddress, allocation_1.Allocation.getPreferredValueAlignment(parameter.type, this.context));
            }
        }
    };
    return FunctionDefinitionBuilder;
}());
exports.FunctionDefinitionBuilder = FunctionDefinitionBuilder;
//# sourceMappingURL=function-definition-builder.js.map