"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var llvm = require("llvm-node");
var allocation_1 = require("../value/allocation");
var types_1 = require("./types");
/**
 * Allocates a llvm array in the entry block of the current function and stores the passed in elements in the array
 * @param elements the elements to be stored in the array
 * @param elementType the type of the elements
 * @param context the context
 * @param name the optional name of the allocation
 * @return the allocation
 */
function allocateLlvmArrayWith(elements, elementType, context, name) {
    var ZERO = llvm.ConstantInt.get(context.llvmContext, 0);
    var arrayType = llvm.ArrayType.get(elementType, elements.length);
    var allocation = allocation_1.Allocation.createAllocaInstInEntryBlock(arrayType, context, name);
    var areAllElementsConstants = elements.every(function (value) { return value instanceof llvm.Constant; });
    var array = context.builder.createInBoundsGEP(allocation, [ZERO, ZERO], name);
    if (areAllElementsConstants) {
        assignFromConstantArray(allocation, elements, elementType, context, name);
    }
    else {
        for (var i = 0; i < elements.length; ++i) {
            var ptr = context.builder.createInBoundsGEP(allocation, [ZERO, llvm.ConstantInt.get(context.llvmContext, i)]);
            context.builder.createAlignedStore(elements[i], ptr, context.module.dataLayout.getPrefTypeAlignment(elementType));
        }
    }
    return array;
}
function assignFromConstantArray(arrayAllocation, elements, elementType, context, name) {
    var arrayType = llvm.ArrayType.get(elementType, elements.length);
    var array = llvm.ConstantArray.get(arrayType, elements);
    var global = new llvm.GlobalVariable(context.module, array.type, true, llvm.LinkageTypes.PrivateLinkage, array, name || "values");
    global.setUnnamedAddr(llvm.UnnamedAddr.Local);
    var pointerType = llvm.Type.getInt8Ty(context.llvmContext).getPointerTo();
    var arrayValue = context.builder.createBitCast(arrayAllocation, pointerType);
    var globalValue = context.builder.createBitCast(global, pointerType);
    var memcpyType = llvm.FunctionType.get(llvm.Type.getVoidTy(context.llvmContext), [
        pointerType,
        pointerType,
        llvm.Type.getInt32Ty(context.llvmContext),
        llvm.Type.getInt32Ty(context.llvmContext),
        llvm.Type.getInt1Ty(context.llvmContext)
    ], false);
    var memcpy = context.module.getOrInsertFunction("llvm.memcpy.p0i8.p0i8.i32", memcpyType);
    context.builder.createCall(memcpy, [
        arrayValue,
        globalValue,
        types_1.sizeof(arrayType, context),
        llvm.ConstantInt.get(context.llvmContext, 0),
        llvm.ConstantInt.getFalse(context.llvmContext)
    ]);
}
function llvmArrayValue(elements, elementType, context, name) {
    if (elements.length === 0) {
        return llvm.ConstantPointerNull.get(elementType.getPointerTo());
    }
    var values;
    if (elements[0] instanceof llvm.Value) {
        values = elements;
    }
    else {
        values = elements.map(function (element) { return context.generateValue(element).generateIR(context); });
    }
    return allocateLlvmArrayWith(values, elementType, context, name);
}
exports.llvmArrayValue = llvmArrayValue;
//# sourceMappingURL=llvm-array-helpers.js.map