"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var llvm = require("llvm-node");
var ts = require("typescript");
var code_generation_diagnostic_1 = require("../../code-generation-diagnostic");
var llvm_array_helpers_1 = require("../util/llvm-array-helpers");
var types_1 = require("../util/types");
var resolved_function_1 = require("./resolved-function");
/**
 * Base class for function references. Handles the coercion of the argument values to the expected types of the function parametes
 * and as well as handling var arg parameters correctly.
 */
var AbstractFunctionReference = (function () {
    /**
     * Creates a new instance
     * @param classType the type of the class to which this function belongs (static or instance method) or absent, if it is
     * a stand alone function.
     */
    function AbstractFunctionReference(classType) {
        this.classType = classType;
    }
    AbstractFunctionReference.prototype.invoke = function (callExpression, callerContext) {
        var resolvedSignature = callerContext.typeChecker.getResolvedSignature(callExpression);
        var resolvedFunction = this.getResolvedFunctionFromSignature(resolvedSignature, callerContext.compilationContext);
        var passedArguments = toLlvmArgumentValues(callExpression.arguments || [], resolvedFunction, callerContext);
        return this.invokeResolvedFunction(resolvedFunction, passedArguments, callerContext);
    };
    AbstractFunctionReference.prototype.invokeWith = function (args, callerContext) {
        return this.invokeResolvedFunction(this.getResolvedFunction(callerContext), args, callerContext);
    };
    AbstractFunctionReference.prototype.getResolvedFunctionFromSignature = function (signature, compilationContext) {
        return resolved_function_1.createResolvedFunctionFromSignature(signature, compilationContext, this.classType);
    };
    AbstractFunctionReference.prototype.invokeResolvedFunction = function (resolvedFunction, args, callerContext) {
        var llvmFunction = this.getLLVMFunction(resolvedFunction, callerContext, args);
        var callArguments = this.getCallArguments(resolvedFunction, args, callerContext);
        var name = resolvedFunction.returnType.flags & ts.TypeFlags.Void ? undefined : resolvedFunction.functionName + "ReturnValue";
        assert(callArguments.length === llvmFunction.getArguments().length, "Calling function with less than expected number of arguments");
        var call = callerContext.builder.createCall(llvmFunction, callArguments, name);
        if (resolvedFunction.returnType.flags & ts.TypeFlags.Void) {
            return;
        }
        else if (resolvedFunction.returnType.flags & ts.TypeFlags.Object) {
            var classReference = callerContext.resolveClass(resolvedFunction.returnType);
            call.addDereferenceableAttr(0, classReference.getTypeStoreSize(resolvedFunction.returnType, callerContext));
        }
        return callerContext.value(call, resolvedFunction.returnType);
    };
    /**
     * Gets the call arguments for invoking the specified function
     * @param resolvedFunction the specific signature of the function to call
     * @param passedArguments the parameters passed in the invoke statement
     * @param callerContext the callers code generation context
     * @return the values that are to be passed to the llvm function
     */
    AbstractFunctionReference.prototype.getCallArguments = function (resolvedFunction, passedArguments, callerContext) {
        var result = [];
        for (var i = 0; i < resolvedFunction.parameters.length; ++i) {
            var parameter = resolvedFunction.parameters[i];
            var arg = void 0;
            if (passedArguments.length > i) {
                arg = passedArguments[i];
            }
            else if (parameter.initializer) {
                arg = callerContext.generateValue(parameter.initializer).generateIR(callerContext);
            }
            else if (parameter.optional) {
                break;
            }
            else if (!parameter.variadic) {
                assert(false, "Missing value for not optional parameter " + parameter.name);
                break;
            }
            if (parameter.variadic) {
                var arrayType = parameter.type;
                var elementType = types_1.getArrayElementType(arrayType);
                result.push(llvm_array_helpers_1.llvmArrayValue(passedArguments.slice(i), types_1.toLLVMType(elementType, callerContext), callerContext, parameter.name), llvm.ConstantInt.get(callerContext.llvmContext, passedArguments.length - i, undefined, false));
                break;
            }
            else {
                result.push(arg);
            }
        }
        return result;
    };
    AbstractFunctionReference.prototype.isAssignable = function () {
        return false;
    };
    AbstractFunctionReference.prototype.isObject = function () {
        return false;
    };
    AbstractFunctionReference.prototype.dereference = function () {
        return this;
    };
    AbstractFunctionReference.prototype.generateIR = function (context) {
        return this.getLLVMFunction(this.getResolvedFunction(context), context);
    };
    AbstractFunctionReference.prototype.castImplicit = function (type, context) {
        // casting functions is not yet supported
        return undefined;
    };
    return AbstractFunctionReference;
}());
exports.AbstractFunctionReference = AbstractFunctionReference;
function toLlvmArgumentValues(args, resolvedFunction, callerContext) {
    var values = [];
    for (var i = 0; i < Math.min(args.length, resolvedFunction.parameters.length); ++i) {
        var parameter = resolvedFunction.parameters[i];
        if (parameter.variadic) {
            var llvmVarArgs = toLlvmVariadicArgument(args.slice(i), parameter, callerContext);
            values.push.apply(values, llvmVarArgs);
        }
        else {
            var argumentValue = toLlvmArgumentValue(args[i], parameter, callerContext);
            values.push(argumentValue);
        }
    }
    return values;
}
function toLlvmVariadicArgument(varArgs, parameter, callerContext) {
    var elementType = types_1.getArrayElementType(parameter.type);
    var llvmVarArgs = new Array(varArgs.length);
    for (var j = 0; j < varArgs.length; ++j) {
        var varArgNode = varArgs[j];
        var castedElement = callerContext.generateValue(varArgNode).castImplicit(elementType, callerContext);
        if (!castedElement) {
            throw code_generation_diagnostic_1.CodeGenerationDiagnostic.unsupportedImplicitCastOfArgument(varArgNode, callerContext.typeChecker.typeToString(elementType), callerContext.typeChecker.typeToString(callerContext.typeChecker.getTypeAtLocation(varArgNode)));
        }
        llvmVarArgs[j] = castedElement.generateIR(callerContext);
    }
    return llvmVarArgs;
}
function toLlvmArgumentValue(arg, parameter, callerContext) {
    var castedElement = callerContext.generateValue(arg).castImplicit(parameter.type, callerContext);
    var argumentType = callerContext.typeChecker.getTypeAtLocation(arg);
    if (!castedElement) {
        throw code_generation_diagnostic_1.CodeGenerationDiagnostic.unsupportedImplicitCastOfArgument(arg, callerContext.typeChecker.typeToString(parameter.type), callerContext.typeChecker.typeToString(argumentType));
    }
    return castedElement.generateIR(callerContext);
}
//# sourceMappingURL=abstract-function-reference.js.map