"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var llvm = require("llvm-node");
var ts = require("typescript");
var types_1 = require("../util/types");
/**
 * Wrapper for an allocation with an alignment.
 *
 * Is capable to work with Values and ensures correct alignment
 */
var Allocation = (function () {
    function Allocation(store, type, alignment, name) {
        this.store = store;
        this.type = type;
        this.alignment = alignment;
        this.name = name;
    }
    Allocation.create = function (type, context, name) {
        var allocaInst = Allocation.createAllocaInstInEntryBlock(types_1.toLLVMType(type, context), context, name);
        var alignment = Allocation.getPreferredValueAlignment(type, context);
        return new Allocation(allocaInst, type, alignment, name);
    };
    Allocation.createAllocaInstInEntryBlock = function (type, context, name) {
        var fn = context.scope.enclosingFunction;
        var entryBlock = fn.getEntryBlock();
        assert(entryBlock, "The function requires an entry block");
        var entryBlockBuilder = new llvm.IRBuilder(fn.getEntryBlock());
        var allocation = entryBlockBuilder.createAlloca(type, undefined, name);
        allocation.alignment = this.getPreferredValueAlignment(type, context);
        return allocation;
    };
    Allocation.createGlobal = function (pointer, type, context, name) {
        assert(pointer.type.isPointerTy(), "Address needs to be a pointer");
        var alignment = Allocation.getPreferredValueAlignment(type, context);
        return new Allocation(pointer, type, alignment, name);
    };
    Allocation.getPreferredValueAlignment = function (type, context) {
        type = type instanceof llvm.Type ? type : types_1.toLLVMType(type, context);
        return context.module.dataLayout.getPrefTypeAlignment(type);
    };
    Allocation.prototype.isAssignable = function () {
        return !(this.store instanceof llvm.GlobalVariable && this.store.constant);
    };
    Allocation.prototype.isObject = function () {
        return false;
    };
    Allocation.prototype.dereference = function (context) {
        if (this.type.flags & ts.TypeFlags.Object) {
            var classReference = context.resolveClass(this.type);
            if (classReference) {
                return classReference.objectFor(new AllocationPointerWrapper(this), this.type, context);
            }
        }
        return context.value(this.generateIR(context), this.type);
    };
    Allocation.prototype.generateIR = function (context) {
        return context.builder.createAlignedLoad(this.store, this.alignment, this.name);
    };
    Allocation.prototype.generateAssignmentIR = function (value, context) {
        assert(this.isAssignable, "Cannot assign to constant global variable");
        var llvmValue = value instanceof llvm.Value ? value : value.generateIR(context);
        context.builder.createAlignedStore(llvmValue, this.store, this.alignment, false);
    };
    Allocation.prototype.castImplicit = function (type, context) {
        return this.dereference(context).castImplicit(type, context);
    };
    return Allocation;
}());
exports.Allocation = Allocation;
/**
 * Wrapper for an allocation that stores an address to use it as pointer
 */
var AllocationPointerWrapper = (function () {
    function AllocationPointerWrapper(allocation) {
        this.allocation = allocation;
    }
    Object.defineProperty(AllocationPointerWrapper.prototype, "type", {
        get: function () {
            return this.allocation.type;
        },
        enumerable: true,
        configurable: true
    });
    AllocationPointerWrapper.prototype.isPointer = function () {
        return true;
    };
    AllocationPointerWrapper.prototype.get = function (context) {
        return this.allocation.generateIR(context);
    };
    AllocationPointerWrapper.prototype.set = function (ptr, context) {
        this.allocation.generateAssignmentIR(ptr, context);
    };
    return AllocationPointerWrapper;
}());
//# sourceMappingURL=allocation.js.map