"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var code_generation_diagnostic_1 = require("../../code-generation-diagnostic");
var types_1 = require("../util/types");
var address_lvalue_1 = require("./address-lvalue");
/**
 * Object reference to a built in object (that is part of the runtime).
 */
var BuiltInObjectReference = (function () {
    /**
     * Creates a new instance for an object that is stored at the specified object address and is of the given type
     * @param objectAddress the address, where the object is stored
     * @param type the type of the object
     * @param clazz the class of the object
     */
    function BuiltInObjectReference(objectAddress, type, clazz) {
        this.objectAddress = objectAddress;
        this.type = type;
        this.clazz = clazz;
        this.properties = new Map();
    }
    BuiltInObjectReference.prototype.getTypeStoreSize = function (context) {
        return this.clazz.getTypeStoreSize(this.type, context);
    };
    BuiltInObjectReference.prototype.generateIR = function (context) {
        return this.objectAddress.get(context);
    };
    BuiltInObjectReference.prototype.isObject = function () {
        return true;
    };
    BuiltInObjectReference.prototype.isAssignable = function () {
        return false;
    };
    BuiltInObjectReference.prototype.dereference = function () {
        return this;
    };
    BuiltInObjectReference.prototype.castImplicit = function (type, context) {
        if (this.type === type || types_1.isMaybeObjectType(type) && type.types.indexOf(this.type) !== -1) {
            return this;
        }
        // casting it to undefined. Casts to other types not yet supported
        if (type.flags & ts.TypeFlags.Undefined) {
            var castedPtr = context.builder.createBitCast(this.generateIR(context), types_1.toLLVMType(type, context));
            return this.clazz.objectFor(new address_lvalue_1.AddressLValue(castedPtr, type), this.type, context);
        }
        return undefined;
    };
    BuiltInObjectReference.prototype.getProperty = function (property, context) {
        var symbol = context.typeChecker.getSymbolAtLocation(property);
        var propertyReference = this.properties.get(symbol);
        if (propertyReference) {
            return propertyReference;
        }
        if (symbol.flags & ts.SymbolFlags.Method) {
            var type = context.typeChecker.getTypeAtLocation(property);
            var apparentType = context.typeChecker.getApparentType(type);
            var signatures = context.typeChecker.getSignaturesOfType(apparentType, ts.SignatureKind.Call);
            var fn = this.createFunctionFor(symbol, signatures, property, context);
            this.properties.set(symbol, fn);
            return fn;
        }
        propertyReference = this.createPropertyReference(symbol, property, context);
        this.properties.set(symbol, propertyReference);
        return propertyReference;
    };
    BuiltInObjectReference.prototype.getIndexer = function (element, context) {
        throw this.throwUnsupportedBuiltIn(element);
    };
    /**
     * Throws an exception for an unsupported index or element accesss
     */
    BuiltInObjectReference.prototype.throwUnsupportedBuiltIn = function (node, symbol) {
        if (node.kind === ts.SyntaxKind.ElementAccessExpression) {
            throw code_generation_diagnostic_1.CodeGenerationDiagnostic.builtInDoesNotSupportElementAccess(node, this.typeName);
        }
        else if (node.kind === ts.SyntaxKind.PropertyAccessExpression) {
            throw code_generation_diagnostic_1.CodeGenerationDiagnostic.builtInPropertyNotSupported(node, this.typeName);
        }
        else {
            throw code_generation_diagnostic_1.CodeGenerationDiagnostic.builtInMethodNotSupported(node, this.typeName, symbol.name);
        }
    };
    /**
     * Creates the function for the given symbol and call
     * @param symbol the symbol of the called function
     * @param signatures the signatures of the method
     * @param propertyAccessExpression the property access
     * @param context the code generation context
     * @return the method to invoke
     * @throws if the built in method is not supported
     */
    BuiltInObjectReference.prototype.createFunctionFor = function (symbol, signatures, propertyAccessExpression, context) {
        return this.throwUnsupportedBuiltIn(propertyAccessExpression);
    };
    /**
     * Creates a property reference for the property defined by the given symbol
     * @param symbol the symbol of the property
     * @param propertyAccess the property access
     * @param context the code generation context
     * @return a property reference for accessing this property
     * @throws if the property is not supported
     */
    BuiltInObjectReference.prototype.createPropertyReference = function (symbol, propertyAccess, context) {
        return this.throwUnsupportedBuiltIn(propertyAccess);
    };
    return BuiltInObjectReference;
}());
exports.BuiltInObjectReference = BuiltInObjectReference;
//# sourceMappingURL=built-in-object-reference.js.map