"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var llvm = require("llvm-node");
var ts = require("typescript");
var types_1 = require("../util/types");
/**
 * Reference to a class (or in JS, to the constructor function)
 */
var ClassReference = (function () {
    /**
     * Creates a new instance
     * @param typeInformation a pointer to the type information of the class
     * @param symbol the symbol for this class
     * @param compilationContext the compilationContext
     */
    function ClassReference(typeInformation, symbol, compilationContext) {
        this.typeInformation = typeInformation;
        this.symbol = symbol;
        this.compilationContext = compilationContext;
        this.llvmType = undefined;
    }
    /**
     * Creates a new type descriptor for the given symbol
     * @param symbol the symbol
     * @param context the context
     * @return a global variable that stores the type descriptor
     */
    ClassReference.createTypeDescriptor = function (symbol, context) {
        var name = symbol.name + "_type_descriptor"; // TODO guarantee uniqueness
        var existing = context.module.getGlobalVariable(name, true);
        if (existing) {
            return existing;
        }
        var nameConstant = llvm.ConstantDataArray.getString(context.llvmContext, symbol.name);
        var nameVariable = new llvm.GlobalVariable(context.module, nameConstant.type, true, llvm.LinkageTypes.PrivateLinkage, nameConstant, symbol.name + "_name");
        nameVariable.setUnnamedAddr(llvm.UnnamedAddr.Global);
        var structType = llvm.StructType.get(context.llvmContext, [nameConstant.type.getPointerTo()], false);
        var struct = llvm.ConstantStruct.get(structType, [nameVariable]);
        return new llvm.GlobalVariable(context.module, struct.type, true, llvm.LinkageTypes.PrivateLinkage, struct, name);
    };
    Object.defineProperty(ClassReference.prototype, "name", {
        get: function () {
            return this.symbol.name;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassReference.prototype, "type", {
        get: function () {
            return this.compilationContext.typeChecker.getDeclaredTypeOfSymbol(this.symbol);
        },
        enumerable: true,
        configurable: true
    });
    ClassReference.prototype.generateIR = function () {
        return this.typeInformation;
    };
    ClassReference.prototype.isObject = function () {
        return false;
    };
    ClassReference.prototype.isAssignable = function () {
        return false;
    };
    ClassReference.prototype.dereference = function () {
        return this;
    };
    ClassReference.prototype.castImplicit = function (type, context) {
        // Can probably never be casted?
        return undefined;
    };
    ClassReference.prototype.getTypeStoreSize = function (objectType, context) {
        var llvmType = this.getLLVMType(objectType, context);
        return context.module.dataLayout.getTypeStoreSize(llvmType);
    };
    ClassReference.prototype.getLLVMType = function (type, context) {
        return this.getObjectType(type, context);
    };
    ClassReference.prototype.getFieldOffset = function (property) {
        var field = this.type.getProperty(property.getName()); // in case the passed in property is transient or this class type is transient
        assert(field.flags & ts.SymbolFlags.Property, "The property " + property.name + " is not a property (e.g. a method instead)");
        var index = this.type.getProperties().indexOf(field);
        assert(index >= 0, "The property " + property.name + " could not be found in class " + this.type.getSymbol().name);
        return index;
    };
    /**
     * Creates the type of the object
     * @param type the type
     * @param context the code generation context
     * @return the llvm type of the object
     */
    ClassReference.prototype.getObjectType = function (type, context) {
        if (!this.llvmType) {
            var forwardDeclaration = llvm.StructType.create(this.compilationContext.llvmContext, "class." + this.symbol.name);
            this.llvmType = forwardDeclaration;
            var fieldTypes = this.getFields(type, context).map(function (field) { return types_1.toLLVMType(field.type, context); });
            if (fieldTypes.length === 0) {
                // LLVM doesn't seem to like empty structs, at least when marked as dereferencaeble (throws value out of range as size is 0).
                // Therefore, add a boolean fake field. Seems to be the same as clang is doing?!
                // TODO this can be removed when a reference to the type descriptor is added to each object
                fieldTypes.push(llvm.Type.getInt1Ty(context.llvmContext));
            }
            forwardDeclaration.setBody(fieldTypes);
        }
        return this.llvmType;
    };
    return ClassReference;
}());
exports.ClassReference = ClassReference;
//# sourceMappingURL=class-reference.js.map