"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var llvm = require("llvm-node");
var ts = require("typescript");
var types_1 = require("../util/types");
/**
 * Factory for creating the declarations of llvm functions
 */
var FunctionFactory = (function () {
    /**
     * Creates a new instance that uses the given name mangler
     * @param nameMangler the name mangler to use
     */
    function FunctionFactory(nameMangler) {
        this.nameMangler = nameMangler;
    }
    /**
     * Creates the declaration (and if needed the definition) for the given resolved function called with the specified number of arguments
     * @param resolvedFunction the function to call (specific function, not overload)
     * @param numberOfArguments the number of arguments passed to the function (to know where optional arguments are unset)
     * @param context the code generation context
     * @param properties properties of the function to create, e.g. is the function readonly, what is the linkage
     * @return {Function} the existing function instance or the newly declared function
     */
    FunctionFactory.prototype.getOrCreate = function (resolvedFunction, numberOfArguments, context, properties) {
        return this.getOrCreateStaticOrInstanceFunction(resolvedFunction, numberOfArguments, context, this.getInitializedProperties(properties));
    };
    /**
     * Creates the declaration (and if needed the definition) for the given resolved function called with the specified number of arguments.
     * The declared function is a instance method that needs to be called with the object instance as first argument
     * @param objectReference the object to which the method belongs
     * @param resolvedFunction the resolved function (not overloaded)
     * @param numberOfArguments the number of arguments passed to the function
     * @param context the code generation context
     * @param properties properties of the function to create, e.g. is the function readonly, what is the linkage
     * @return {Function} the existing or newly declared function
     */
    FunctionFactory.prototype.getOrCreateInstanceMethod = function (objectReference, resolvedFunction, numberOfArguments, context, properties) {
        assert(resolvedFunction.instanceMethod && resolvedFunction.classType, "Resolved function needs to be an instance method");
        return this.getOrCreateStaticOrInstanceFunction(resolvedFunction, numberOfArguments, context, this.getInitializedProperties(properties), objectReference);
    };
    FunctionFactory.prototype.getInitializedProperties = function (properties) {
        return Object.assign({}, this.getDefaultFunctionProperties(), properties);
    };
    FunctionFactory.prototype.getDefaultFunctionProperties = function () {
        return {
            linkage: llvm.LinkageTypes.LinkOnceODRLinkage,
            readonly: false,
            readnone: false,
            alwaysInline: false,
            noInline: false,
            noUnwind: false,
            visibility: llvm.VisibilityTypes.Default
        };
    };
    FunctionFactory.prototype.getOrCreateStaticOrInstanceFunction = function (resolvedFunction, numberOfArguments, context, properties, objectReference) {
        var mangledName = this.getMangledFunctionName(resolvedFunction, numberOfArguments);
        var fn = context.module.getFunction(mangledName);
        if (!fn) {
            fn = this.createFunction(mangledName, resolvedFunction, numberOfArguments, context, properties, objectReference);
        }
        return fn;
    };
    FunctionFactory.prototype.createFunction = function (mangledName, resolvedFunction, numberOfArguments, context, properties, objectReference) {
        var llvmArgumentTypes = toLlvmArgumentTypes(resolvedFunction, numberOfArguments, context, objectReference);
        var functionType = llvm.FunctionType.get(types_1.toLLVMType(resolvedFunction.returnType, context), llvmArgumentTypes, false);
        var fn = llvm.Function.create(functionType, properties.linkage, mangledName, context.module);
        fn.visibility = properties.visibility;
        for (var _i = 0, _a = this.getFunctionAttributes(properties); _i < _a.length; _i++) {
            var attribute = _a[_i];
            fn.addFnAttr(attribute);
        }
        this.attributeParameters(fn, resolvedFunction, context, objectReference);
        if (resolvedFunction.returnType.flags & ts.TypeFlags.Object) {
            var classReference = context.resolveClass(resolvedFunction.returnType);
            // If object can be undefined, or null
            fn.addDereferenceableAttr(0, classReference.getTypeStoreSize(resolvedFunction.returnType, context));
        }
        return fn;
    };
    FunctionFactory.prototype.getFunctionAttributes = function (properties) {
        var attributes = [];
        if (properties.noUnwind) {
            attributes.push(llvm.Attribute.AttrKind.NoUnwind);
        }
        if (properties.alwaysInline) {
            attributes.push(llvm.Attribute.AttrKind.AlwaysInline);
        }
        if (properties.noInline) {
            attributes.push(llvm.Attribute.AttrKind.NoInline);
        }
        if (properties.readonly) {
            attributes.push(llvm.Attribute.AttrKind.ReadOnly);
        }
        if (properties.readnone) {
            attributes.push(llvm.Attribute.AttrKind.ReadNone);
        }
        return attributes;
    };
    FunctionFactory.prototype.attributeParameters = function (fn, resolvedFunction, context, object) {
        var parameters = fn.getArguments();
        var argumentOffset = 0;
        if (object) {
            var self = parameters[0];
            self.addAttr(llvm.Attribute.AttrKind.ReadOnly);
            self.addDereferenceableAttr(object.getTypeStoreSize(context));
            argumentOffset = 1;
        }
        for (var i = argumentOffset; i < parameters.length; ++i) {
            var parameter = parameters[i];
            var parameterDefinition = resolvedFunction.parameters[i - argumentOffset];
            if (parameterDefinition.variadic) {
                break;
            }
            if (parameterDefinition.type.flags & ts.TypeFlags.Object) {
                var classReference = context.resolveClass(parameterDefinition.type);
                parameter.addDereferenceableAttr(classReference.getTypeStoreSize(parameterDefinition.type, context));
            }
        }
    };
    FunctionFactory.prototype.getMangledFunctionName = function (resolvedFunction, numberOfArguments) {
        var typesOfUsedParameters = [];
        for (var i = 0; i < resolvedFunction.parameters.length; ++i) {
            var parameter = resolvedFunction.parameters[i];
            if (parameter.optional && !parameter.initializer && numberOfArguments <= i) {
                break; // Optional parameter that is not set. Therefore, this parameter is not actually used
            }
            typesOfUsedParameters.push(parameter.type);
        }
        return this.mangleFunctionName(resolvedFunction, typesOfUsedParameters);
    };
    FunctionFactory.prototype.mangleFunctionName = function (resolvedFunction, typesOfUsedParameters) {
        if (resolvedFunction.classType) {
            return this.nameMangler.mangleMethodName(resolvedFunction.classType, resolvedFunction.functionName, typesOfUsedParameters, resolvedFunction.sourceFile);
        }
        return this.nameMangler.mangleFunctionName(resolvedFunction.functionName, typesOfUsedParameters, resolvedFunction.sourceFile);
    };
    return FunctionFactory;
}());
exports.FunctionFactory = FunctionFactory;
function toLlvmArgumentTypes(resolvedFunction, numberOfArguments, context, objectReference) {
    var argumentTypes = objectReference ? [types_1.toLLVMType(objectReference.type, context)] : [];
    for (var i = 0; i < resolvedFunction.parameters.length; ++i) {
        var parameter = resolvedFunction.parameters[i];
        if (parameter.variadic) {
            var elementType = parameter.type.typeArguments[0];
            argumentTypes.push(types_1.toLLVMType(elementType, context).getPointerTo(), llvm.Type.getInt32Ty(context.llvmContext));
            break;
        }
        else if (parameter.optional && !parameter.initializer && numberOfArguments <= i) {
            // optional argument that is not set, skip
            break;
        }
        else {
            argumentTypes.push(types_1.toLLVMType(parameter.type, context));
        }
    }
    return argumentTypes;
}
//# sourceMappingURL=function-factory.js.map