"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var llvm = require("llvm-node");
var allocation_1 = require("./allocation");
/**
 * Reference to an object property (x.name)
 */
var ObjectPropertyReference = (function () {
    function ObjectPropertyReference(propertyType, object) {
        this.propertyType = propertyType;
        this.object = object;
    }
    /**
     * Creates a computed property
     * @param propertyType the type of the property
     * @param object the object to which the property belongs
     * @param property the property symbol
     * @param getter the getter to access the property
     * @param setter the setter to change the value of the property
     * @return the property reference
     */
    ObjectPropertyReference.createComputedPropertyReference = function (propertyType, object, property, getter, setter) {
        return new ComputedObjectPropertyReference(propertyType, object, property, getter, setter);
    };
    ObjectPropertyReference.createFieldProperty = function (propertyType, object, property) {
        return new ObjectFieldPropertyReference(propertyType, object, property);
    };
    ObjectPropertyReference.prototype.generateIR = function (context) {
        return this.getValue(context).generateIR(context);
    };
    ObjectPropertyReference.prototype.generateAssignmentIR = function (value, context) {
        this.setValue(value, context);
    };
    ObjectPropertyReference.prototype.isObject = function () {
        return false;
    };
    ObjectPropertyReference.prototype.dereference = function (context) {
        return this.getValue(context);
    };
    ObjectPropertyReference.prototype.castImplicit = function (type, context) {
        return this.dereference(context).castImplicit(type, context);
    };
    return ObjectPropertyReference;
}());
exports.ObjectPropertyReference = ObjectPropertyReference;
/**
 * Backs a computed property, e.g. one that is implemented using get index() or one that resits in the runtime
 */
var ComputedObjectPropertyReference = (function (_super) {
    __extends(ComputedObjectPropertyReference, _super);
    function ComputedObjectPropertyReference(propertyType, object, property, getter, setter) {
        var _this = _super.call(this, propertyType, object) || this;
        _this.property = property;
        _this.getter = getter;
        _this.setter = setter;
        return _this;
    }
    ComputedObjectPropertyReference.prototype.getValue = function (context) {
        var result = context.builder.createCall(this.getter, [this.object.generateIR(context)], this.property.name);
        return context.value(result, this.propertyType);
    };
    ComputedObjectPropertyReference.prototype.setValue = function (value, context) {
        var args = [
            value.generateIR(context),
            this.object.generateIR(context)
        ].reverse();
        context.builder.createCall(this.setter, args);
    };
    ComputedObjectPropertyReference.prototype.isAssignable = function () {
        return !!this.setter;
    };
    return ComputedObjectPropertyReference;
}(ObjectPropertyReference));
/**
 * Backs a property that dispatches directly to an object field
 */
var ObjectFieldPropertyReference = (function (_super) {
    __extends(ObjectFieldPropertyReference, _super);
    function ObjectFieldPropertyReference(propertyType, object, property) {
        var _this = _super.call(this, propertyType, object) || this;
        _this.property = property;
        return _this;
    }
    ObjectFieldPropertyReference.prototype.getValue = function (context) {
        var alignment = allocation_1.Allocation.getPreferredValueAlignment(this.propertyType, context);
        var value = context.builder.createAlignedLoad(this.getFieldAddress(context), alignment, this.property.name);
        return context.value(value, this.propertyType);
    };
    ObjectFieldPropertyReference.prototype.setValue = function (value, context) {
        var alignment = allocation_1.Allocation.getPreferredValueAlignment(this.propertyType, context);
        context.builder.createAlignedStore(value.generateIR(context), this.getFieldAddress(context), alignment);
    };
    ObjectFieldPropertyReference.prototype.isAssignable = function () {
        return true;
    };
    ObjectFieldPropertyReference.prototype.getFieldAddress = function (context) {
        var fieldIndex = [
            llvm.ConstantInt.get(context.llvmContext, 0),
            llvm.ConstantInt.get(context.llvmContext, this.object.clazz.getFieldOffset(this.property))
        ];
        return context.builder.createInBoundsGEP(this.object.generateIR(context), fieldIndex, "&" + this.property.name);
    };
    return ObjectFieldPropertyReference;
}(ObjectPropertyReference));
//# sourceMappingURL=object-property-reference.js.map