"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var llvm = require("llvm-node");
var ts = require("typescript");
var function_declaration_builder_1 = require("../util/function-declaration-builder");
var types_1 = require("../util/types");
/**
 * Represents a primitive, stack allocated value.
 */
var Primitive = (function () {
    function Primitive(llvmValue, type) {
        this.llvmValue = llvmValue;
        this.type = type;
    }
    /**
     * Converts the value to an int32 type
     * @param value the value to convert
     * @param valueType the type of the value
     * @param int32Type the int 32 type
     * @param context the context
     * @return the converted value
     * @see https://tc39.github.io/ecma262/#sec-ecmascript-function-objects-call-thisargument-argumentslist
     */
    Primitive.toInt32 = function (value, valueType, int32Type, context) {
        var intValue;
        if (valueType.flags & ts.TypeFlags.IntLike) {
            intValue = value.generateIR(context);
        }
        else if (valueType.flags & ts.TypeFlags.BooleanLike) {
            var llvmValue = value.generateIR(context);
            intValue = context.builder.createZExt(llvmValue, llvm.Type.getInt32Ty(context.llvmContext), llvmValue.name + "AsInt32");
        }
        else if (valueType.flags & (ts.TypeFlags.NumberLike)) {
            var llvmValue = value.generateIR(context);
            var fn = function_declaration_builder_1.FunctionDeclarationBuilder.create("toInt32d", [{
                    name: "value",
                    type: valueType
                }], int32Type, context)
                .withAttribute(llvm.Attribute.AttrKind.ReadNone)
                .withAttribute(llvm.Attribute.AttrKind.AlwaysInline)
                .externalLinkage()
                .declareIfNotExisting();
            intValue = context.builder.createCall(fn, [llvmValue], llvmValue.name + "AsInt32");
        }
        else if (valueType.flags & ts.TypeFlags.Object || types_1.isMaybeObjectType(valueType)) {
            intValue = llvm.ConstantInt.get(context.llvmContext, 0); // as long as valueOf and toString are not supported
        }
        else {
            throw new Error("Unsupported conversion of " + context.typeChecker.typeToString(valueType) + " to int32");
        }
        return new Primitive(intValue, int32Type);
    };
    /**
     * Converts the given value to a boolean
     * @param value the value to convert
     * @param valueType the type of the value
     * @param context the context
     * @return a boolean value
     */
    Primitive.toBoolean = function (value, valueType, context) {
        var llvmValue = value instanceof llvm.Value ? value : value.generateIR(context);
        if (valueType.flags & ts.TypeFlags.BooleanLike) {
            return llvmValue;
        }
        if (valueType.flags & ts.TypeFlags.IntLike) {
            return context.builder.createICmpNE(llvmValue, llvm.ConstantInt.get(context.llvmContext, 0), llvmValue.name + "AsBool");
        }
        if (valueType.flags & ts.TypeFlags.NumberLike) {
            return context.builder.createFCmpONE(llvmValue, llvm.ConstantFP.get(context.llvmContext, 0), llvmValue.name + "AsBool");
        }
        if (valueType.flags & ts.TypeFlags.Undefined) {
            return llvm.ConstantInt.getFalse(context.llvmContext);
        }
        if (valueType.flags & ts.TypeFlags.Object || types_1.isMaybeObjectType(valueType)) {
            return context.builder.createICmpNE(llvmValue, llvm.Constant.getNullValue(llvmValue.type), llvmValue.name + "AsBool");
        }
        throw new Error("value of type " + context.typeChecker.typeToString(valueType) + " cannot be converted to bool");
    };
    /**
     * Converts the value to a number
     * @param value the value to convert
     * @param valueType the type of the value
     * @param numberType the typescript 'number' type
     * @param context the context
     * @return {Primitive} the value converted to a number
     */
    Primitive.toNumber = function (value, valueType, numberType, context) {
        var numberValue;
        var llvmValue = value.generateIR(context);
        if (valueType.flags & ts.TypeFlags.BooleanLike) {
            numberValue = context.builder.createUIToFP(llvmValue, llvm.Type.getDoubleTy(context.llvmContext), llvmValue.name + "AsNumber");
        }
        else if (valueType.flags & ts.TypeFlags.IntLike) {
            numberValue = context.builder.createSIToFP(llvmValue, llvm.Type.getDoubleTy(context.llvmContext), llvmValue.name + "AsNumber");
        }
        else if (valueType.flags & ts.TypeFlags.NumberLike) {
            numberValue = llvmValue;
        }
        else if (valueType.flags & ts.TypeFlags.Undefined) {
            numberValue = llvm.ConstantFP.getNaN(llvm.Type.getDoubleTy(context.llvmContext));
        }
        else if (valueType.flags & ts.TypeFlags.Object || types_1.isMaybeObjectType(valueType)) {
            numberValue = llvm.ConstantFP.get(context.llvmContext, 0);
        }
        else {
            throw new Error("value of type " + llvmValue.type + " cannot be converted to number");
        }
        return new Primitive(numberValue, numberType);
    };
    Primitive.prototype.generateIR = function () {
        return this.llvmValue;
    };
    Primitive.prototype.isAssignable = function () {
        return false;
    };
    Primitive.prototype.isObject = function () {
        return false;
    };
    Primitive.prototype.dereference = function () {
        return this;
    };
    Primitive.prototype.castImplicit = function (type, context) {
        if (this.type === type) {
            return this;
        }
        if (type.flags & ts.TypeFlags.BooleanLike) {
            // no safe implicit cast to boolean, skip
            return undefined;
        }
        if (type.flags & ts.TypeFlags.IntLike) {
            // only boolean can safely be casted to int
            return this.type.flags & ts.TypeFlags.BooleanLike ? Primitive.toInt32(this, this.type, type, context) : undefined;
        }
        // all primitive types can safely be casted to number
        if (type.flags & ts.TypeFlags.NumberLike) {
            return Primitive.toNumber(this, this.type, type, context);
        }
        return undefined;
    };
    return Primitive;
}());
exports.Primitive = Primitive;
//# sourceMappingURL=primitive.js.map