"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var ts = require("typescript");
/**
 * Creates a new resolved function
 */
function createResolvedFunction(name, typeParameters, parameters, returnType, sourceFile, classType, instanceMethod, async) {
    if (instanceMethod === void 0) { instanceMethod = false; }
    if (async === void 0) { async = false; }
    return {
        async: async,
        functionName: name,
        parameters: parameters,
        typeParameters: typeParameters,
        instanceMethod: instanceMethod,
        returnType: returnType,
        sourceFile: sourceFile,
        classType: classType
    };
}
exports.createResolvedFunction = createResolvedFunction;
/**
 * Creates a new resolved parameter
 */
function createResolvedParameter(name, type, optional, initializer, variadic) {
    if (optional === void 0) { optional = false; }
    if (variadic === void 0) { variadic = false; }
    return {
        name: name,
        type: type,
        optional: optional,
        initializer: initializer,
        variadic: variadic
    };
}
exports.createResolvedParameter = createResolvedParameter;
/**
 * Creates a new resolved function from a signature
 * @param signature the signature
 * @param compilationContext the compilation context
 * @param classType the class type if the function belongs to a class
 * @returns the created resolved function
 */
function createResolvedFunctionFromSignature(signature, compilationContext, classType) {
    var returnType = signature.getReturnType();
    var definition = signature.declaration.body ? signature.declaration : undefined;
    var symbol = signature.declaration.name ? compilationContext.typeChecker.getSymbolAtLocation(signature.declaration.name) : undefined;
    if (symbol && symbol.declarations && symbol.declarations.length > 1) {
        var isOverloadPredicate = function (decl) { return compilationContext.typeChecker.isImplementationOfOverload(decl); };
        definition = symbol.declarations.find(isOverloadPredicate) || definition;
    }
    return {
        async: !!signature.declaration.modifiers && !!signature.declaration.modifiers.find(function (modifier) { return modifier.kind === ts.SyntaxKind.AsyncKeyword; }),
        classType: classType || getClassType(signature, compilationContext.typeChecker),
        declaration: signature.declaration,
        definition: definition,
        functionName: getDeclaredFunctionName(signature.declaration, compilationContext.typeChecker),
        instanceMethod: isInstanceMethod(signature.declaration),
        parameters: getResolvedParameters(signature, compilationContext.typeChecker),
        returnType: returnType,
        sourceFile: signature.declaration.getSourceFile(),
        symbol: symbol,
        typeParameters: signature.typeParameters || []
    };
}
exports.createResolvedFunctionFromSignature = createResolvedFunctionFromSignature;
function getClassType(signature, typeChecker) {
    var declaration = signature.getDeclaration();
    if (declaration.kind === ts.SyntaxKind.ConstructSignature) {
        return signature.getReturnType();
    }
    if (declaration.kind === ts.SyntaxKind.MethodSignature) {
        return typeChecker.getTypeAtLocation(declaration.parent);
    }
    return undefined;
}
function isInstanceMethod(declaration) {
    if (declaration.kind === ts.SyntaxKind.ConstructSignature) {
        return false;
    }
    if (declaration.kind === ts.SyntaxKind.MethodSignature || declaration.kind === ts.SyntaxKind.MethodDeclaration) {
        var modifiers = declaration.modifiers || [];
        return !modifiers.find(function (modifier) { return modifier.kind === ts.SyntaxKind.StaticKeyword; });
    }
    return false;
}
function getDeclaredFunctionName(declaration, typeChecker) {
    if (declaration.kind === ts.SyntaxKind.ConstructSignature || declaration.kind === ts.SyntaxKind.Constructor) {
        return "constructor";
    }
    else {
        assert(declaration.name, "Anonymous functions are not supported");
        return typeChecker.getSymbolAtLocation(declaration.name).name;
    }
}
function getResolvedParameters(signature, typeChecker) {
    var declaration = signature.getDeclaration();
    var parameters = [];
    for (var i = 0; i < declaration.parameters.length; ++i) {
        var parameter = declaration.parameters[i];
        var parameterSymbol = signature.parameters[i];
        var parameterType = typeChecker.getTypeOfSymbolAtLocation(parameterSymbol, parameter);
        parameters.push({
            initializer: parameter.initializer,
            name: parameterSymbol.name,
            optional: !!parameter.questionToken,
            type: parameterType,
            symbol: parameterSymbol,
            variadic: !!parameter.dotDotDotToken
        });
    }
    return parameters;
}
//# sourceMappingURL=resolved-function.js.map