"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var llvm = require("llvm-node");
var ts = require("typescript");
var default_name_mangler_1 = require("../default-name-mangler");
var function_declaration_builder_1 = require("../util/function-declaration-builder");
var function_definition_builder_1 = require("../util/function-definition-builder");
var types_1 = require("../util/types");
var abstract_function_reference_1 = require("./abstract-function-reference");
var address_lvalue_1 = require("./address-lvalue");
var resolved_function_1 = require("./resolved-function");
var speedyjs_function_factory_1 = require("./speedyjs-function-factory");
var speedyjs_object_reference_1 = require("./speedyjs-object-reference");
var SpeedyJSConstructorFunctionReference = (function (_super) {
    __extends(SpeedyJSConstructorFunctionReference, _super);
    function SpeedyJSConstructorFunctionReference(resolvedFunction, fn) {
        var _this = _super.call(this, resolvedFunction.classType) || this;
        _this.resolvedFunction = resolvedFunction;
        _this.fn = fn;
        return _this;
    }
    SpeedyJSConstructorFunctionReference.create = function (signature, classReference, context) {
        var resolvedFunction;
        if (signature.declaration) {
            resolvedFunction = resolved_function_1.createResolvedFunctionFromSignature(signature, context.compilationContext, classReference.type);
            speedyjs_function_factory_1.verifyIsSupportedSpeedyJSFunction(signature.declaration, context);
        }
        else {
            // default constructor
            var sourceFile = classReference.type.getSymbol().declarations[0].getSourceFile();
            resolvedFunction = resolved_function_1.createResolvedFunction("constructor", [], [], signature.getReturnType(), sourceFile, classReference.type);
        }
        var declarationContext = context.createChildContext();
        var fn = new ConstructorFunctionBuilder(resolvedFunction, classReference, declarationContext)
            .mangleName()
            .defineIfAbsent();
        return new SpeedyJSConstructorFunctionReference(resolvedFunction, fn);
    };
    SpeedyJSConstructorFunctionReference.prototype.getResolvedFunction = function (callerContext) {
        return this.resolvedFunction;
    };
    SpeedyJSConstructorFunctionReference.prototype.getResolvedFunctionFromSignature = function (signature, compilationContext) {
        return signature.declaration ? _super.prototype.getResolvedFunctionFromSignature.call(this, signature, compilationContext) : this.resolvedFunction;
    };
    SpeedyJSConstructorFunctionReference.prototype.getLLVMFunction = function () {
        return this.fn;
    };
    return SpeedyJSConstructorFunctionReference;
}(abstract_function_reference_1.AbstractFunctionReference));
exports.SpeedyJSConstructorFunctionReference = SpeedyJSConstructorFunctionReference;
var ConstructorFunctionBuilder = (function () {
    function ConstructorFunctionBuilder(resolvedFunction, classReference, context) {
        this.resolvedFunction = resolvedFunction;
        this.classReference = classReference;
        this.context = context;
        this.constructorName = resolvedFunction.functionName;
    }
    ConstructorFunctionBuilder.prototype.mangleName = function () {
        var argumentTypes = this.resolvedFunction.parameters.map(function (parameter) { return parameter.type; });
        var nameMangler = new default_name_mangler_1.DefaultNameMangler(this.context.compilationContext);
        this.constructorName = nameMangler.mangleMethodName(this.classReference.type, this.constructorName, argumentTypes, this.resolvedFunction.sourceFile);
        return this;
    };
    ConstructorFunctionBuilder.prototype.defineIfAbsent = function () {
        var existingFn = this.context.module.getFunction(this.constructorName);
        if (existingFn) {
            return existingFn;
        }
        return this.define();
    };
    ConstructorFunctionBuilder.prototype.define = function () {
        var declaration = function_declaration_builder_1.FunctionDeclarationBuilder
            .forResolvedFunction(this.resolvedFunction, this.context)
            .linkOnceOdrLinkage()
            .name(this.constructorName)
            .declare();
        var entryBlock = llvm.BasicBlock.create(this.context.llvmContext, "entry", declaration);
        this.context.builder.setInsertionPoint(entryBlock);
        var objectAddress = this.allocateObjectOnHeap();
        var objectReference = new speedyjs_object_reference_1.SpeedyJSObjectReference(objectAddress, this.classReference.type, this.classReference);
        this.context.enterChildScope();
        this.context.scope.addVariable(this.classReference.symbol, objectReference);
        this.initializeFields(objectAddress);
        this.callUserConstructorFn(declaration, objectReference);
        this.context.leaveChildScope();
        return declaration;
    };
    ConstructorFunctionBuilder.prototype.allocateObjectOnHeap = function () {
        var objectType = this.classReference.getLLVMType(this.classReference.type, this.context);
        var pointerType = llvm.Type.getInt8PtrTy(this.context.llvmContext);
        var mallocFunctionType = llvm.FunctionType.get(pointerType, [llvm.Type.getInt32Ty(this.context.llvmContext)], false);
        var malloc = this.context.module.getOrInsertFunction("malloc", mallocFunctionType);
        var size = types_1.sizeof(objectType, this.context);
        var result = this.context.builder.createCall(malloc, [size], "thisVoid*");
        var ptr = this.context.builder.createBitCast(result, objectType.getPointerTo(), "this");
        return new address_lvalue_1.AddressLValue(ptr, this.classReference.type);
    };
    ConstructorFunctionBuilder.prototype.initializeFields = function (objectAddress) {
        var fields = this.classReference.type.getApparentProperties().filter(function (property) { return property.flags & ts.SymbolFlags.Property; });
        for (var _i = 0, fields_1 = fields; _i < fields_1.length; _i++) {
            var field = fields_1[_i];
            var declaration = field.valueDeclaration;
            var value = void 0;
            if (declaration.initializer) {
                value = this.context.generateValue(declaration.initializer).generateIR(this.context);
            }
            else if (this.context.compilationContext.compilerOptions.unsafe) {
                continue;
            }
            else {
                value = llvm.Constant.getNullValue(types_1.toLLVMType(this.context.typeChecker.getTypeAtLocation(declaration), this.context));
            }
            var fieldOffset = llvm.ConstantInt.get(this.context.llvmContext, this.classReference.getFieldOffset(field));
            var gepFieldIndex = [llvm.ConstantInt.get(this.context.llvmContext, 0), fieldOffset];
            var fieldPointer = this.context.builder.createInBoundsGEP(objectAddress.get(this.context), gepFieldIndex, "&" + field.name);
            this.context.builder.createStore(value, fieldPointer, false);
        }
    };
    ConstructorFunctionBuilder.prototype.callUserConstructorFn = function (fn, objectReference) {
        if (!this.resolvedFunction.definition) {
            this.context.builder.createRet(objectReference.generateIR(this.context));
            return;
        }
        function_definition_builder_1.FunctionDefinitionBuilder.create(fn, this.resolvedFunction, this.context)
            .returnValue(objectReference)
            .self(objectReference)
            .define();
    };
    return ConstructorFunctionBuilder;
}());
//# sourceMappingURL=speedyjs-constructor-function-reference.js.map