"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var debug = require("debug");
var llvm = require("llvm-node");
var ts = require("typescript");
var built_in_symbols_1 = require("./built-in-symbols");
var code_generation_diagnostic_1 = require("./code-generation-diagnostic");
var default_code_generation_context_factory_1 = require("./code-generation/default-code-generation-context-factory");
var not_yet_implemented_code_generator_1 = require("./code-generation/not-yet-implemented-code-generator");
var per_file_code_generator_1 = require("./code-generation/per-file/per-file-code-generator");
var log_unknown_transform_visitor_1 = require("./transform/log-unknown-transform-visitor");
var speedyjs_transform_visitor_1 = require("./transform/speedyjs-transform-visitor");
var transform_visitor_1 = require("./transform/transform-visitor");
var typescript_type_checker_1 = require("./typescript-type-checker");
var LOG = debug("compiler");
/**
 * The main Interface for compiling a SpeedyJS Program.
 * The compiler creates compiles all with "use speedyjs" marked functions and rewrites the original source file
 * to invoke the web assembly function instead of executing the function body in the JS runtime.
 */
var Compiler = (function () {
    /**
     * Creates a new instance that uses the given compilation options and compiler host
     * @param compilerOptions the compilation options
     * @param compilerHost the compiler host
     */
    function Compiler(compilerOptions, compilerHost) {
        this.compilerOptions = compilerOptions;
        this.compilerHost = compilerHost;
    }
    /**
     * Compiles the source files with the given names
     * @param rootFileNames the file names of the source files to compile
     * @return the result of the compilation.
     */
    Compiler.prototype.compile = function (rootFileNames) {
        LOG("Start Compiling");
        Compiler.initLLVM();
        var program = ts.createProgram(rootFileNames, this.compilerOptions, this.compilerHost);
        var diagnostics = program.getSyntacticDiagnostics().concat(program.getOptionsDiagnostics(), program.getGlobalDiagnostics(), program.getSemanticDiagnostics());
        if (diagnostics.length > 0) {
            return { diagnostics: diagnostics, exitStatus: ts.ExitStatus.DiagnosticsPresent_OutputsSkipped };
        }
        var context = new llvm.LLVMContext();
        var builtIns = built_in_symbols_1.BuiltInSymbols.create(program, this.compilerHost);
        var compilationContext = {
            builtIns: builtIns,
            compilerHost: this.compilerHost,
            compilerOptions: this.compilerOptions,
            llvmContext: context,
            typeChecker: new typescript_type_checker_1.TypeScriptTypeChecker(program.getTypeChecker()),
            rootDir: program.getCommonSourceDirectory()
        };
        return Compiler.emit(program, compilationContext);
    };
    Compiler.initLLVM = function () {
        llvm.initializeAllTargets();
        llvm.initializeAllTargetInfos();
        llvm.initializeAllAsmPrinters();
        llvm.initializeAllTargetMCs();
        llvm.initializeAllAsmParsers();
    };
    Compiler.emit = function (program, compilationContext) {
        var codeGenerationContextFactory = new default_code_generation_context_factory_1.DefaultCodeGenerationContextFactory(new not_yet_implemented_code_generator_1.NotYetImplementedCodeGenerator());
        var codeGenerator = new per_file_code_generator_1.PerFileCodeGenerator(compilationContext.llvmContext, codeGenerationContextFactory);
        var logUnknownVisitor = new log_unknown_transform_visitor_1.LogUnknownTransformVisitor();
        var speedyJSVisitor = new speedyjs_transform_visitor_1.SpeedyJSTransformVisitor(compilationContext, codeGenerator);
        try {
            var emitResult = program.emit(undefined, undefined, undefined, undefined, { before: [
                    transform_visitor_1.createTransformVisitorFactory(logUnknownVisitor),
                    transform_visitor_1.createTransformVisitorFactory(speedyJSVisitor)
                ] });
            var exitStatus = void 0;
            if (emitResult.emitSkipped) {
                exitStatus = ts.ExitStatus.DiagnosticsPresent_OutputsSkipped;
            }
            else {
                codeGenerator.completeCompilation();
                exitStatus = ts.ExitStatus.Success;
            }
            LOG("Program compiled");
            return { exitStatus: exitStatus, diagnostics: emitResult.diagnostics };
        }
        catch (ex) {
            if (ex instanceof code_generation_diagnostic_1.CodeGenerationDiagnostic) {
                return { exitStatus: ts.ExitStatus.DiagnosticsPresent_OutputsSkipped, diagnostics: [ex.toDiagnostic()] };
            }
            else {
                LOG("Compilation failed", ex);
                throw ex;
            }
        }
    };
    return Compiler;
}());
exports.Compiler = Compiler;
//# sourceMappingURL=compiler.js.map