"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var assert = require("assert");
var debug = require("debug");
var fs = require("fs");
var path = require("path");
var speedyjs_runtime_1 = require("speedyjs-runtime");
var ts = require("typescript");
var llvm_nm_1 = require("./llvm-nm");
var tools_1 = require("./tools");
var EXECUTABLE_NAME = "llvm-link";
var LOG = debug("external-tools/llvm-link");
/**
 * Linker to link multiple llvm files into a single output file.
 * The implementation determines automatically which files are to be included by the
 * yet unresolved symbols
 */
var LLVMLink = (function () {
    /**
     * @param buildDirectory the build directory
     */
    function LLVMLink(buildDirectory) {
        this.buildDirectory = buildDirectory;
        this.byteCodeFiles = [];
        this.byteCodeSymbolResolver = new llvm_nm_1.LLVMByteCodeSymbolsResolver();
    }
    /**
     * Adds a file that is to be linked
     * @param file the file to be linked
     */
    LLVMLink.prototype.addByteCodeFile = function (file) {
        assert(file && ts.sys.fileExists(file), "The file " + file + " to link does not exist.");
        this.byteCodeFiles.push(file);
    };
    /**
     * Adds the files needed by the runtime
     * @param unsafe should the unsafe runtime (without safe memory guarantees) be used
     */
    LLVMLink.prototype.addRuntime = function (unsafe) {
        if (unsafe === void 0) { unsafe = false; }
        if (unsafe) {
            this.addByteCodeFile(speedyjs_runtime_1.UNSAFE_RUNTIME);
        }
        else {
            this.addByteCodeFile(speedyjs_runtime_1.SAFE_RUNTIME);
        }
    };
    /**
     * Adds the shared libs (e.g. c(++) std lib) to the linkage
     */
    LLVMLink.prototype.addSharedLibs = function () {
        this.addObjectFilesFromDirectory(speedyjs_runtime_1.SHARED_LIBRARIES_DIRECTORY);
    };
    LLVMLink.prototype.addObjectFilesFromDirectory = function (directory) {
        for (var _i = 0, _a = fs.readdirSync(directory); _i < _a.length; _i++) {
            var file = _a[_i];
            var fullPath = path.join(directory, file);
            if (fullPath === speedyjs_runtime_1.COMPILER_RT_FILE) {
                continue;
            }
            if (file.endsWith(".a")) {
                (_b = this.byteCodeFiles).push.apply(_b, this.extractArchive(fullPath));
            }
            else if (file.endsWith(".bc") || file.endsWith(".o")) {
                this.byteCodeFiles.push(fullPath);
            }
        }
        var _b;
    };
    /**
     * Links the files together
     * @param target the path to the file where the linked output should be written to
     * @param entrySymbols the name of the external visible entry symbols
     */
    LLVMLink.prototype.link = function (target, entrySymbols) {
        var linkingFiles = Array.from(this.getObjectFilesToLink(this.byteCodeFiles, entrySymbols));
        var quotedFileNames = Array.from(linkingFiles).map(function (file) { return "\"" + file + "\""; }).join(" ");
        LOG(tools_1.execLLVM(EXECUTABLE_NAME, quotedFileNames + " -o \"" + target + "\""));
        return target;
    };
    LLVMLink.prototype.extractArchive = function (archiveFilePath) {
        var directory = this.buildDirectory.getTempSubdirectory(path.basename(archiveFilePath));
        LOG("Extract archive " + archiveFilePath + " to " + directory);
        tools_1.execLLVM("llvm-ar", "x \"" + archiveFilePath + "\"", directory);
        return fs.readdirSync(directory).map(function (file) { return path.join(directory, file); });
    };
    LLVMLink.prototype.getObjectFilesToLink = function (objectFiles, entrySymbols) {
        var loopAgain = true;
        var includedObjectFiles = new Set();
        var unresolvedSymbols = new Set(entrySymbols);
        while (loopAgain) {
            loopAgain = false;
            for (var _i = 0, objectFiles_1 = objectFiles; _i < objectFiles_1.length; _i++) {
                var objectFile = objectFiles_1[_i];
                if (includedObjectFiles.has(objectFile)) {
                    continue;
                }
                if (this.considerObjectFile(objectFile, unresolvedSymbols)) {
                    includedObjectFiles.add(objectFile);
                    loopAgain = true;
                }
            }
        }
        return includedObjectFiles.values();
    };
    LLVMLink.prototype.considerObjectFile = function (objectFile, unresolvedSymbols) {
        var symbols = this.byteCodeSymbolResolver.getSymbols(objectFile);
        LOG("Consider object file " + objectFile + " for linkage");
        if (intersects(unresolvedSymbols, symbols.defined)) {
            LOG("Add object file " + objectFile + " to linkage");
            symbols.defined.forEach(function (symbol) { return unresolvedSymbols.delete(symbol); });
            symbols.undefined.forEach(function (symbol) { return unresolvedSymbols.add(symbol); });
            return true;
        }
        return false;
    };
    return LLVMLink;
}());
exports.LLVMLink = LLVMLink;
/**
 * Helper function that tests if the two sets intersect (there is no such function in the API!!!!)
 * Is O(n) :/
 */
function intersects(setA, setB) {
    var smaller = setA.size < setB.size ? setA : setB;
    var larger = setA === smaller ? setB : setA;
    for (var _i = 0, _a = Array.from(smaller); _i < _a.length; _i++) {
        var symbol = _a[_i];
        if (larger.has(symbol)) {
            return true;
        }
    }
    return false;
}
//# sourceMappingURL=llvm-link.js.map