"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
/**
 * Creates a factory for a ts Transformer that calls the function of a transform visitor.
 * @param transformVisitor the transform visitor to return
 * @return the factory
 */
function createTransformVisitorFactory(transformVisitor) {
    return function (transformationContext) {
        return createTransformer(transformVisitor, transformationContext);
    };
}
exports.createTransformVisitorFactory = createTransformVisitorFactory;
/**
 * Creates a transformer for the given transform visitor
 * @param transformVisitor the transform visitor to wrap
 * @param transformationContext the transformation context
 * @return the transformer that wrapps the transform visitor
 */
function createTransformer(transformVisitor, transformationContext) {
    var context = {
        requestEmitHelper: function (emitHelper) {
            transformationContext.requestEmitHelper(emitHelper);
        },
        visit: function (node, visitor) {
            visitor = visitor || transformVisitor;
            var kindName = ts.SyntaxKind[node.kind];
            var propertyName = "visit" + kindName;
            var visitorFn = visitor[propertyName];
            if (visitorFn) {
                return visitorFn.call(transformVisitor, node, context);
            }
            else {
                return visitor.fallback(node, context);
            }
        },
        visitEachChild: function (node, visitor) {
            return ts.visitEachChild(node, function (child) { return context.visit(child, visitor); }, transformationContext);
        },
        visitLexicalEnvironment: function (nodes, visitor) {
            return ts.visitLexicalEnvironment(nodes, function (child) { return context.visit(child, visitor); }, transformationContext);
        }
    };
    return context.visit;
}
exports.createTransformer = createTransformer;
//# sourceMappingURL=transform-visitor.js.map