"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
/**
 * Wrapper of the type script type checker.
 * It mainly gets rid of all nullable types. If nullable types are supported, than the unwrapping
 * of nullable types should be removed.
 */
var TypeScriptTypeChecker = (function () {
    function TypeScriptTypeChecker(tsTypeChecker) {
        this.tsTypeChecker = tsTypeChecker;
    }
    TypeScriptTypeChecker.prototype.getAliasedSymbol = function (symbol) {
        return this.tsTypeChecker.getAliasedSymbol(symbol);
    };
    TypeScriptTypeChecker.prototype.getApparentType = function (type) {
        return this.toSupportedType(this.tsTypeChecker.getApparentType(type));
    };
    TypeScriptTypeChecker.prototype.getSignaturesOfType = function (type, kind) {
        return this.tsTypeChecker.getSignaturesOfType(type, kind);
    };
    TypeScriptTypeChecker.prototype.getReturnTypeOfSignature = function (signature) {
        return this.toSupportedType(this.tsTypeChecker.getReturnTypeOfSignature(signature));
    };
    TypeScriptTypeChecker.prototype.getDeclaredTypeOfSymbol = function (symbol) {
        return this.toSupportedType(this.tsTypeChecker.getDeclaredTypeOfSymbol(symbol));
    };
    TypeScriptTypeChecker.prototype.getSymbolAtLocation = function (node) {
        return this.tsTypeChecker.getSymbolAtLocation(node);
    };
    TypeScriptTypeChecker.prototype.getFullyQualifiedName = function (symbol) {
        return this.tsTypeChecker.getFullyQualifiedName(symbol);
    };
    TypeScriptTypeChecker.prototype.getSignatureFromDeclaration = function (functionDeclaration) {
        return new SignatureWrapper(this.tsTypeChecker.getSignatureFromDeclaration(functionDeclaration), this.tsTypeChecker);
    };
    TypeScriptTypeChecker.prototype.getTypeAtLocation = function (node) {
        var type = this.toSupportedType(this.tsTypeChecker.getTypeAtLocation(node));
        // e.g. when const x: int[] = [] then the type of [] is never[] that is quite unfortunate. Take the contextual
        // type information into consideration in this case (but do not otherwise. Otherwise let x: number = 3 returns unexpected results.
        if (type.flags & ts.TypeFlags.Object && type.objectFlags & ts.ObjectFlags.Reference) {
            var typeReference = type;
            if (typeReference.typeArguments && typeReference.typeArguments.some(function (typeArgument) { return !!(typeArgument.flags & ts.TypeFlags.Never); })) {
                type = this.toSupportedType(this.getContextualType(node)) || type;
            }
        }
        return type;
    };
    TypeScriptTypeChecker.prototype.getContextualType = function (node) {
        return this.toSupportedType(this.tsTypeChecker.getContextualType(node));
    };
    TypeScriptTypeChecker.prototype.typeToString = function (type) {
        return this.tsTypeChecker.typeToString(type);
    };
    TypeScriptTypeChecker.prototype.getTypeOfSymbolAtLocation = function (symbol, location) {
        return this.toSupportedType(this.tsTypeChecker.getTypeOfSymbolAtLocation(symbol, location));
    };
    TypeScriptTypeChecker.prototype.getResolvedSignature = function (callLikeExpression) {
        return new SignatureWrapper(this.tsTypeChecker.getResolvedSignature(callLikeExpression), this.tsTypeChecker);
    };
    TypeScriptTypeChecker.prototype.isImplementationOfOverload = function (fun) {
        return this.tsTypeChecker.isImplementationOfOverload(fun);
    };
    TypeScriptTypeChecker.prototype.toSupportedType = function (type) {
        return toSupportedType(type, this.tsTypeChecker);
    };
    TypeScriptTypeChecker.prototype.isUndefinedSymbol = function (symbol) {
        return this.tsTypeChecker.isUndefinedSymbol(symbol);
    };
    return TypeScriptTypeChecker;
}());
exports.TypeScriptTypeChecker = TypeScriptTypeChecker;
function toSupportedType(type, typeChecker) {
    // should never happen but it does!, thanks typescript
    if (typeof (type) === "undefined") {
        return type;
    }
    // e.g. 1, 2... we are not interested in the literals, only in the type
    if (type.flags & ts.TypeFlags.Literal) {
        return typeChecker.getBaseTypeOfLiteralType(type);
    }
    // e.g. 1 | 2. We are not interested in the actual literals, just what the type is.
    if (type.flags & ts.TypeFlags.Union) {
        var unionType = type;
        var intLiterals = unionType.types.every(function (t) { return !!(t.flags & ts.TypeFlags.IntLike); });
        var numberLiterals = unionType.types.every(function (t) { return !!(t.flags & ts.TypeFlags.NumberLike) && !(t.flags & ts.TypeFlags.IntLike); });
        var booleanLiterals = unionType.types.every(function (t) { return !!(t.flags & ts.TypeFlags.BooleanLike); });
        if ((intLiterals || numberLiterals || booleanLiterals) && unionType.types.length > 0) {
            return toSupportedType(unionType.types[0], typeChecker);
        }
    }
    if (type.flags & ts.TypeFlags.Union) {
        var unionType = type;
        var isMaybeType = unionType.types.length === 2 && unionType.types.some(function (t) { return !!(t.flags & ts.TypeFlags.Undefined); });
        // e.g. double | undefined, int | undefined, boolean | undefined are not supported. Compiler does not allow undefined values of
        // primitive types (yet). Maybe in the future when union types are fully supported. So until then, just return the non nullable type
        if (isMaybeType && unionType.types.some(function (t) { return !!(t.flags & (ts.TypeFlags.IntLike | ts.TypeFlags.NumberLike | ts.TypeFlags.BooleanLike)); })) {
            return unionType.getNonNullableType();
        }
    }
    return type;
}
var SignatureWrapper = (function () {
    function SignatureWrapper(signature, typeChecker) {
        this.signature = signature;
        this.typeChecker = typeChecker;
    }
    SignatureWrapper.prototype.getJsDocTags = function () {
        return this.signature.getJsDocTags();
    };
    Object.defineProperty(SignatureWrapper.prototype, "declaration", {
        get: function () {
            return this.signature.declaration;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SignatureWrapper.prototype, "typeParameters", {
        get: function () {
            return this.signature.typeParameters;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SignatureWrapper.prototype, "parameters", {
        get: function () {
            return this.signature.parameters;
        },
        enumerable: true,
        configurable: true
    });
    SignatureWrapper.prototype.getDeclaration = function () {
        return this.signature.getDeclaration();
    };
    SignatureWrapper.prototype.getTypeParameters = function () {
        return this.signature.getTypeParameters();
    };
    SignatureWrapper.prototype.getParameters = function () {
        return this.signature.getParameters();
    };
    SignatureWrapper.prototype.getReturnType = function () {
        var returnType = this.signature.getReturnType();
        // if is Promise<T>
        if (returnType.getSymbol() &&
            returnType.getSymbol().getName() === "Promise" &&
            returnType.flags & ts.TypeFlags.Object &&
            returnType.objectFlags & ts.ObjectFlags.Reference) {
            var typeReference = returnType;
            returnType = typeReference.typeArguments && typeReference.typeArguments.length === 1 ? typeReference.typeArguments[0] : returnType;
        }
        return toSupportedType(returnType, this.typeChecker);
    };
    SignatureWrapper.prototype.getDocumentationComment = function () {
        return this.signature.getDocumentationComment();
    };
    return SignatureWrapper;
}());
//# sourceMappingURL=typescript-type-checker.js.map